"""Tests for matcher.py."""

import textwrap

from pytype import config
from pytype import file_utils
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.tests import test_base
from pytype.tests import test_utils

import unittest


class MatcherTestBase(test_base.UnitTest):

  def setUp(self):
    super().setUp()
    options = config.Options.create(python_version=self.python_version)
    self.ctx = test_utils.make_context(options)
    self.matcher = self.ctx.matcher(self.ctx.root_node)

  def _match_var(self, left, right):
    var = self.ctx.program.NewVariable()
    var.AddBinding(left, [], self.ctx.root_node)
    for view in abstract_utils.get_views([var], self.ctx.root_node):
      yield self.matcher.match_var_against_type(var, right, {}, view)

  def assertMatch(self, left, right):
    for match in self._match_var(left, right):
      self.assertEqual(match, {})

  def assertNoMatch(self, left, right):
    for match in self._match_var(left, right):
      self.assertIsNone(match)


class MatcherTest(MatcherTestBase):
  """Test matcher.AbstractMatcher."""

  def _make_class(self, name):
    return abstract.InterpreterClass(name, [], {}, None, self.ctx)

  def _parse_and_lookup(self, src, objname, filename=None):
    if filename is None:
      filename = str(hash(src))
    with file_utils.Tempdir() as d:
      d.create_file(filename + ".pyi", src)
      self.ctx.options.tweak(pythonpath=[d.path])  # monkeypatch
      ast = self.ctx.loader.import_name(filename)
      return ast.Lookup(filename + "." + objname)

  def _convert(self, x, name, as_instance=False):
    pyval = self._parse_and_lookup(x, name)
    if as_instance:
      pyval = abstract_utils.AsInstance(pyval)
    return self.ctx.convert.constant_to_value(pyval, {}, self.ctx.root_node)

  def _convert_type(self, t, as_instance=False):
    """Convenience function for turning a string into an abstract value.

    Note that this function cannot be called more than once per test with
    the same arguments, since we hash the arguments to get a filename for
    the temporary pyi.

    Args:
      t: The string representation of a type.
      as_instance: Whether to convert as an instance.

    Returns:
      A BaseValue.
    """
    src = textwrap.dedent(f"""
      from typing import Any, Callable, Iterator, Tuple, Type, Union
      from protocols import Sequence, SupportsLower
      x = ...  # type: {t}
    """)
    filename = str(hash((t, as_instance)))
    x = self._parse_and_lookup(src, "x", filename).type
    if as_instance:
      x = abstract_utils.AsInstance(x)
    return self.ctx.convert.constant_to_value(x, {}, self.ctx.root_node)

  def test_basic(self):
    self.assertMatch(abstract.Empty(self.ctx), abstract.Empty(self.ctx))

  def test_type(self):
    left = self._make_class("dummy")
    type_parameters = {
        abstract_utils.T: abstract.TypeParameter(abstract_utils.T, self.ctx)
    }
    other_type = abstract.ParameterizedClass(self.ctx.convert.type_type,
                                             type_parameters, self.ctx)
    for result in self._match_var(left, other_type):
      instance_binding, = result[abstract_utils.T].bindings
      self.assertEqual(instance_binding.data.cls, left)

  def test_union(self):
    left_option1 = self._make_class("o1")
    left_option2 = self._make_class("o2")
    left = abstract.Union([left_option1, left_option2], self.ctx)
    self.assertMatch(left, self.ctx.convert.type_type)

  def test_metaclass(self):
    left = self._make_class("left")
    meta1 = self._make_class("m1")
    meta2 = self._make_class("m2")
    left.set_class(
        self.ctx.root_node,
        self.ctx.program.NewVariable([meta1, meta2], [], self.ctx.root_node))
    self.assertMatch(left, meta1)
    self.assertMatch(left, meta2)

  def test_empty_against_class(self):
    var = self.ctx.program.NewVariable()
    right = self._make_class("bar")
    result = self.matcher.match_var_against_type(var, right, {}, {})
    self.assertEqual(result, {})

  def test_empty_var_against_empty(self):
    var = self.ctx.program.NewVariable()
    right = abstract.Empty(self.ctx)
    result = self.matcher.match_var_against_type(var, right, {}, {})
    self.assertEqual(result, {})

  def test_empty_against_type_parameter(self):
    var = self.ctx.program.NewVariable()
    right = abstract.TypeParameter("T", self.ctx)
    result = self.matcher.match_var_against_type(var, right, {}, {})
    self.assertCountEqual(result.keys(), ["T"])
    self.assertFalse(result["T"].bindings)

  def test_empty_against_unsolvable(self):
    var = self.ctx.program.NewVariable()
    right = abstract.Empty(self.ctx)
    result = self.matcher.match_var_against_type(var, right, {}, {})
    self.assertEqual(result, {})

  def test_class_against_type_union(self):
    left = self._make_class("foo")
    union = abstract.Union((left,), self.ctx)
    right = abstract.ParameterizedClass(self.ctx.convert.type_type,
                                        {abstract_utils.T: union}, self.ctx)
    self.assertMatch(left, right)

  def test_none_against_bool(self):
    # See pep484.COMPAT_ITEMS.
    left = self._convert_type("None", as_instance=True)
    right = self._convert_type("bool")
    self.assertMatch(left, right)

  def test_homogeneous_tuple(self):
    left = self._convert_type("Tuple[int, ...]", as_instance=True)
    right1 = self._convert_type("Tuple[int, ...]")
    right2 = self._convert_type("Tuple[str, ...]")
    self.assertMatch(left, right1)
    self.assertNoMatch(left, right2)

  def test_heterogeneous_tuple(self):
    left1 = self._convert_type("Tuple[Union[int, str]]", as_instance=True)
    left2 = self._convert_type("Tuple[int, str]", as_instance=True)
    left3 = self._convert_type("Tuple[str, int]", as_instance=True)
    right = self._convert_type("Tuple[int, str]")
    self.assertNoMatch(left1, right)
    self.assertMatch(left2, right)
    self.assertNoMatch(left3, right)

  def test_heterogeneous_tuple_against_homogeneous_tuple(self):
    left = self._convert_type("Tuple[bool, int]", as_instance=True)
    right1 = self._convert_type("Tuple[bool, ...]")
    right2 = self._convert_type("Tuple[int, ...]")
    right3 = self._convert_type("tuple")
    self.assertNoMatch(left, right1)
    self.assertMatch(left, right2)
    self.assertMatch(left, right3)

  def test_homogeneous_tuple_against_heterogeneous_tuple(self):
    left1 = self._convert_type("Tuple[bool, ...]", as_instance=True)
    left2 = self._convert_type("Tuple[int, ...]", as_instance=True)
    left3 = self._convert_type("tuple", as_instance=True)
    right = self._convert_type("Tuple[bool, int]")
    self.assertMatch(left1, right)
    self.assertNoMatch(left2, right)
    self.assertMatch(left3, right)

  def test_tuple_type(self):
    # homogeneous against homogeneous
    left = self._convert_type("Type[Tuple[float, ...]]", as_instance=True)
    right1 = self._convert_type("Type[Tuple[float, ...]]")
    right2 = self._convert_type("Type[Tuple[str, ...]]")
    self.assertMatch(left, right1)
    self.assertNoMatch(left, right2)

    # heterogeneous against heterogeneous
    left1 = self._convert_type("Type[Tuple[Union[int, str]]]", as_instance=True)
    left2 = self._convert_type("Type[Tuple[int, str]]", as_instance=True)
    left3 = self._convert_type("Type[Tuple[str, int]]", as_instance=True)
    right = self._convert_type("Type[Tuple[int, str]]")
    self.assertNoMatch(left1, right)
    self.assertMatch(left2, right)
    self.assertNoMatch(left3, right)

    # heterogeneous against homogeneous
    left = self._convert_type("Type[Tuple[bool, int]]", as_instance=True)
    right1 = self._convert_type("Type[Tuple[bool, ...]]")
    right2 = self._convert_type("Type[Tuple[int, ...]]")
    right3 = self._convert_type("Type[tuple]")
    self.assertNoMatch(left, right1)
    self.assertMatch(left, right2)
    self.assertMatch(left, right3)

    # homogeneous against heterogeneous
    left1 = self._convert_type("Type[Tuple[bool, ...]]", as_instance=True)
    left2 = self._convert_type("Type[Tuple[int, ...]]", as_instance=True)
    left3 = self._convert_type("Type[tuple]", as_instance=True)
    right = self._convert_type("Type[Tuple[bool, int]]")
    self.assertMatch(left1, right)
    self.assertNoMatch(left2, right)
    self.assertMatch(left3, right)

  def test_tuple_subclass(self):
    left = self._convert("""
      from typing import Tuple
      class A(Tuple[bool, int]): ...""", "A", as_instance=True)
    right1 = self._convert_type("Tuple[bool, int]")
    right2 = self._convert_type("Tuple[int, bool]")
    right3 = self._convert_type("Tuple[int, int]")
    right4 = self._convert_type("Tuple[int]")
    right5 = self._convert_type("tuple")
    right6 = self._convert_type("Tuple[bool, ...]")
    right7 = self._convert_type("Tuple[int, ...]")
    self.assertMatch(left, right1)
    self.assertNoMatch(left, right2)
    self.assertMatch(left, right3)
    self.assertNoMatch(left, right4)
    self.assertMatch(left, right5)
    self.assertNoMatch(left, right6)
    self.assertMatch(left, right7)

  def test_annotation_class(self):
    left = abstract.AnnotationClass("Dict", self.ctx)
    right = self.ctx.convert.object_type
    self.assertMatch(left, right)

  def test_empty_tuple_class(self):
    var = self.ctx.program.NewVariable()
    params = {
        0: abstract.TypeParameter(abstract_utils.K, self.ctx),
        1: abstract.TypeParameter(abstract_utils.V, self.ctx)
    }
    params[abstract_utils.T] = abstract.Union((params[0], params[1]), self.ctx)
    right = abstract.TupleClass(self.ctx.convert.tuple_type, params, self.ctx)
    match = self.matcher.match_var_against_type(var, right, {}, {})
    self.assertSetEqual(set(match), {abstract_utils.K, abstract_utils.V})

  def test_unsolvable_against_tuple_class(self):
    left = self.ctx.convert.unsolvable
    params = {
        0: abstract.TypeParameter(abstract_utils.K, self.ctx),
        1: abstract.TypeParameter(abstract_utils.V, self.ctx)
    }
    params[abstract_utils.T] = abstract.Union((params[0], params[1]), self.ctx)
    right = abstract.TupleClass(self.ctx.convert.tuple_type, params, self.ctx)
    for match in self._match_var(left, right):
      self.assertSetEqual(set(match), {abstract_utils.K, abstract_utils.V})
      self.assertEqual(match[abstract_utils.K].data,
                       [self.ctx.convert.unsolvable])
      self.assertEqual(match[abstract_utils.V].data,
                       [self.ctx.convert.unsolvable])

  def test_bool_against_float(self):
    left = self.ctx.convert.true
    right = self.ctx.convert.primitive_classes[float]
    self.assertMatch(left, right)

  def test_pytd_function_against_callable(self):
    f = self._convert("def f(x: int) -> bool: ...", "f")
    plain_callable = self._convert_type("Callable")
    good_callable1 = self._convert_type("Callable[[bool], int]")
    good_callable2 = self._convert_type("Callable[..., int]")
    self.assertMatch(f, plain_callable)
    self.assertMatch(f, good_callable1)
    self.assertMatch(f, good_callable2)

  def test_pytd_function_against_callable_bad_return(self):
    f = self._convert("def f(x: int) -> bool: ...", "f")
    callable_bad_ret = self._convert_type("Callable[[int], str]")
    self.assertNoMatch(f, callable_bad_ret)

  def test_pytd_function_against_callable_bad_arg_count(self):
    f = self._convert("def f(x: int) -> bool: ...", "f")
    callable_bad_count1 = self._convert_type("Callable[[], bool]")
    callable_bad_count2 = self._convert_type("Callable[[int, str], bool]")
    self.assertNoMatch(f, callable_bad_count1)
    self.assertNoMatch(f, callable_bad_count2)

  def test_pytd_function_against_callable_bad_arg_type(self):
    f = self._convert("def f(x: bool) -> bool: ...", "f")
    callable_bad_arg1 = self._convert_type("Callable[[int], bool]")
    callable_bad_arg2 = self._convert_type("Callable[[str], bool]")
    self.assertNoMatch(f, callable_bad_arg1)
    self.assertNoMatch(f, callable_bad_arg2)

  def test_bound_pytd_function_against_callable(self):
    instance = self._convert("""
      class A:
        def f(self, x: int) -> bool: ...
    """, "A", as_instance=True)
    binding = instance.to_binding(self.ctx.root_node)
    _, var = self.ctx.attribute_handler.get_attribute(self.ctx.root_node,
                                                      instance, "f", binding)
    bound = var.data[0]
    _, var = self.ctx.attribute_handler.get_attribute(self.ctx.root_node,
                                                      instance.cls, "f")
    unbound = var.data[0]
    callable_no_self = self._convert_type("Callable[[int]]")
    callable_self = self._convert_type("Callable[[Any, int]]")
    self.assertMatch(bound, callable_no_self)
    self.assertNoMatch(unbound, callable_no_self)
    self.assertNoMatch(bound, callable_self)
    self.assertMatch(unbound, callable_self)

  def test_native_function_against_callable(self):
    # Matching a native function against a callable always succeeds, regardless
    # of argument and return types.
    f = abstract.NativeFunction("f", lambda x: x, self.ctx)
    callable_type = self._convert_type("Callable[[int], int]")
    self.assertMatch(f, callable_type)

  def test_callable_instance(self):
    left1 = self._convert_type("Callable[[int], bool]", as_instance=True)
    left2 = self._convert_type("Callable", as_instance=True)
    left3 = self._convert_type("Callable[..., int]", as_instance=True)
    right1 = self._convert_type("Callable[[bool], int]")
    right2 = self._convert_type("Callable[..., int]")
    right3 = self._convert_type("Callable")
    self.assertMatch(left1, right1)
    self.assertMatch(left2, right1)
    self.assertMatch(left3, right1)
    self.assertMatch(left1, right2)
    self.assertMatch(left2, right2)
    self.assertMatch(left3, right2)
    self.assertMatch(left1, right3)
    self.assertMatch(left2, right3)
    self.assertMatch(left3, right3)

  def test_callable_instance_bad_return(self):
    left1 = self._convert_type("Callable[[int], float]", as_instance=True)
    left2 = self._convert_type("Callable[..., float]", as_instance=True)
    right1 = self._convert_type("Callable[[bool], int]")
    right2 = self._convert_type("Callable[..., int]")
    self.assertNoMatch(left1, right1)
    self.assertNoMatch(left2, right1)
    self.assertNoMatch(left1, right2)
    self.assertNoMatch(left2, right2)

  def test_callable_instance_bad_arg_count(self):
    left1 = self._convert_type("Callable[[], int]", as_instance=True)
    left2 = self._convert_type("Callable[[str, str], int]", as_instance=True)
    right = self._convert_type("Callable[[str], int]")
    self.assertNoMatch(left1, right)
    self.assertNoMatch(left2, right)

  def test_callable_instance_bad_arg_type(self):
    left1 = self._convert_type("Callable[[bool], Any]", as_instance=True)
    left2 = self._convert_type("Callable[[str], Any]", as_instance=True)
    right = self._convert_type("Callable[[int], Any]")
    self.assertNoMatch(left1, right)
    self.assertNoMatch(left2, right)

  def test_type_against_callable(self):
    left1 = self._convert_type("Type[int]", as_instance=True)
    left2 = self._convert_type("Type[str]", as_instance=True)
    right1 = self._convert_type("Callable[..., float]")
    right2 = self._convert_type("Callable[[], float]")
    self.assertMatch(left1, right1)
    self.assertMatch(left1, right2)
    self.assertNoMatch(left2, right1)
    self.assertNoMatch(left2, right2)

  def test_anystr_instance_against_anystr(self):
    right = self.ctx.convert.name_to_value("typing.AnyStr")
    dummy_instance = abstract.Instance(self.ctx.convert.tuple_type, self.ctx)
    left = abstract.TypeParameterInstance(right, dummy_instance, self.ctx)
    for result in self._match_var(left, right):
      self.assertCountEqual(
          [(name, var.data) for name, var in result.items()],
          [("typing.AnyStr", [left])])

  def test_protocol(self):
    left1 = self._convert_type("str", as_instance=True)
    left2 = self._convert("""
      class A:
        def lower(self) : ...
    """, "A", as_instance=True)
    left3 = self._convert_type("int", as_instance=True)
    right = self._convert_type("SupportsLower")
    self.assertMatch(left1, right)
    self.assertMatch(left2, right)
    self.assertNoMatch(left3, right)

  def test_protocol_iterator(self):
    left1 = self._convert_type("Iterator", as_instance=True)
    left2 = self._convert("""
      class A:
        def __next__(self): ...
        def __iter__(self): ...
    """, "A", as_instance=True)
    left3 = self._convert_type("int", as_instance=True)
    right = self._convert_type("Iterator")
    self.assertMatch(left1, right)
    self.assertMatch(left2, right)
    self.assertNoMatch(left3, right)

  def test_protocol_sequence(self):
    left1 = self._convert_type("list", as_instance=True)
    left2 = self._convert("""
      class A:
        def __getitem__(self, i) : ...
        def __len__(self): ...
    """, "A", as_instance=True)
    left3 = self._convert_type("int", as_instance=True)
    right = self._convert_type("Sequence")
    self.assertMatch(left1, right)
    self.assertMatch(left2, right)
    self.assertNoMatch(left3, right)

  @unittest.skip("Needs to be fixed, tries to match protocol against A")
  def test_parameterized_protocol(self):
    left1 = self._convert("""
      from typing import Iterator
      class A:
        def __iter__(self) -> Iterator[int] : ...
    """, "A", as_instance=True)
    left2 = self._convert_type("int", as_instance=True)
    right = self._convert_type("Iterable[int]")
    self.assertMatch(left1, right)
    self.assertNoMatch(left2, right)

  def test_no_return(self):
    self.assertMatch(self.ctx.convert.no_return, self.ctx.convert.no_return)

  def test_empty_against_no_return(self):
    self.assertNoMatch(self.ctx.convert.empty, self.ctx.convert.no_return)

  def test_no_return_against_class(self):
    right = self._convert_type("int")
    self.assertNoMatch(self.ctx.convert.no_return, right)

  def test_empty_against_parameterized_iterable(self):
    left = self.ctx.convert.empty
    right = abstract.ParameterizedClass(
        self.ctx.convert.list_type,
        {abstract_utils.T: abstract.TypeParameter(abstract_utils.T, self.ctx)},
        self.ctx)
    for subst in self._match_var(left, right):
      self.assertSetEqual(set(subst), {abstract_utils.T})
      self.assertListEqual(subst[abstract_utils.T].data,
                           [self.ctx.convert.empty])

  def test_list_against_mapping(self):
    left = self._convert_type("list", as_instance=True)
    right = self.ctx.convert.name_to_value("typing.Mapping")
    self.assertNoMatch(left, right)

  def test_list_against_parameterized_mapping(self):
    left = self._convert_type("list", as_instance=True)
    right = abstract.ParameterizedClass(
        self.ctx.convert.name_to_value("typing.Mapping"), {
            abstract_utils.K:
                abstract.TypeParameter(abstract_utils.K, self.ctx),
            abstract_utils.V:
                abstract.TypeParameter(abstract_utils.V, self.ctx)
        }, self.ctx)
    self.assertNoMatch(left, right)


class TypeVarTest(MatcherTestBase):
  """Test matching TypeVar against various types."""

  def test_match_from_mro(self):
    # A TypeParameter never matches anything in match_from_mro, since its mro is
    # empty. This test is mostly to make sure we don't crash.
    self.assertIsNone(
        self.matcher.match_from_mro(
            abstract.TypeParameter("T", self.ctx), self.ctx.convert.int_type))

  def test_compute_subst(self):
    formal_args = [("x", self.ctx.convert.unsolvable)]
    x_val = abstract.TypeParameter("T", self.ctx)
    arg_dict = {"x": x_val.to_binding(self.ctx.root_node)}
    view = {arg_dict["x"].variable: arg_dict["x"]}
    subst, bad_param = self.matcher.compute_subst(formal_args, arg_dict, view)
    self.assertEqual(subst, {})
    self.assertIsNone(bad_param)

  def test_compute_subst_no_match(self):
    formal_args = [("x", self.ctx.convert.int_type)]
    x_val = abstract.TypeParameter("T", self.ctx)
    arg_dict = {"x": x_val.to_binding(self.ctx.root_node)}
    view = {arg_dict["x"].variable: arg_dict["x"]}
    subst, bad_param = self.matcher.compute_subst(formal_args, arg_dict, view)
    self.assertIsNone(subst)
    self.assertEqual(bad_param.name, "x")

  def test_bad_matches(self):
    self.assertFalse(
        self.matcher.bad_matches(
            abstract.TypeParameter("T",
                                   self.ctx).to_variable(self.ctx.root_node),
            self.ctx.convert.unsolvable)[0])

  def test_bad_matches_no_match(self):
    self.assertTrue(
        self.matcher.bad_matches(
            abstract.TypeParameter("T",
                                   self.ctx).to_variable(self.ctx.root_node),
            self.ctx.convert.int_type)[0])

  def test_any(self):
    self.assertMatch(
        abstract.TypeParameter("T", self.ctx), self.ctx.convert.unsolvable)

  def test_object(self):
    self.assertMatch(
        abstract.TypeParameter("T", self.ctx), self.ctx.convert.object_type)

  def test_type(self):
    self.assertMatch(
        abstract.TypeParameter("T", self.ctx), self.ctx.convert.type_type)

  def test_parameterized_type(self):
    self.assertMatch(
        abstract.TypeParameter("T", self.ctx),
        abstract.ParameterizedClass(
            self.ctx.convert.type_type,
            {abstract_utils.T: self.ctx.convert.unsolvable}, self.ctx))

  def test_parameterized_type_no_match(self):
    self.assertNoMatch(
        abstract.TypeParameter("T", self.ctx),
        abstract.ParameterizedClass(
            self.ctx.convert.type_type,
            {abstract_utils.T: self.ctx.convert.int_type}, self.ctx))

  def test_nested(self):
    self.assertMatch(
        abstract.ParameterizedClass(
            self.ctx.convert.list_type,
            {abstract_utils.T: abstract.TypeParameter("T", self.ctx)},
            self.ctx), self.ctx.convert.type_type)

  def test_nested_no_match(self):
    self.assertNoMatch(
        abstract.ParameterizedClass(
            self.ctx.convert.list_type,
            {abstract_utils.T: abstract.TypeParameter("T", self.ctx)},
            self.ctx), self.ctx.convert.list_type)

  def test_no_match(self):
    self.assertNoMatch(
        abstract.TypeParameter("T", self.ctx), self.ctx.convert.int_type)


if __name__ == "__main__":
  unittest.main()



from typing import Any

class classproperty:
    im_func: Any
    def __init__(self, func) -> None: ...
    def __get__(self, obj, cls): ...
    @property
    def __func__(self): ...

class hybrid_method:
    func: Any
    def __init__(self, func) -> None: ...
    def __get__(self, obj, cls): ...

def memoize_single_value(func): ...

class memoized_property:
    __func__: Any
    __name__: Any
    __doc__: Any
    def __init__(self, func) -> None: ...
    def __get__(self, obj, cls): ...
    def clear_cache(self, obj) -> None: ...
    def peek_cache(self, obj, default: Any | None = ...): ...

def deprecated_function(
    msg: Any | None = ...,
    deprecated: Any | None = ...,
    removed: Any | None = ...,
    updoc: bool = ...,
    replacement: Any | None = ...,
    _is_method: bool = ...,
    func_module: Any | None = ...,
): ...
def deprecated_method(
    msg: Any | None = ...,
    deprecated: Any | None = ...,
    removed: Any | None = ...,
    updoc: bool = ...,
    replacement: Any | None = ...,
): ...



from typing import Any

from .Image import ImagePointHandler

DESCRIPTION: str
VERSION: str
core: Any
INTENT_PERCEPTUAL: int
INTENT_RELATIVE_COLORIMETRIC: int
INTENT_SATURATION: int
INTENT_ABSOLUTE_COLORIMETRIC: int
DIRECTION_INPUT: int
DIRECTION_OUTPUT: int
DIRECTION_PROOF: int
FLAGS: Any

class ImageCmsProfile:
    def __init__(self, profile) -> None: ...
    def tobytes(self): ...

class ImageCmsTransform(ImagePointHandler):
    transform: Any
    input_mode: Any
    output_mode: Any
    output_profile: Any
    def __init__(
        self, input, output, input_mode, output_mode, intent=..., proof: Any | None = ..., proof_intent=..., flags: int = ...
    ) -> None: ...
    def point(self, im): ...
    def apply(self, im, imOut: Any | None = ...): ...
    def apply_in_place(self, im): ...

def get_display_profile(handle: Any | None = ...): ...

class PyCMSError(Exception): ...

def profileToProfile(
    im, inputProfile, outputProfile, renderingIntent=..., outputMode: Any | None = ..., inPlace: bool = ..., flags: int = ...
): ...
def getOpenProfile(profileFilename): ...
def buildTransform(inputProfile, outputProfile, inMode, outMode, renderingIntent=..., flags: int = ...): ...
def buildProofTransform(
    inputProfile, outputProfile, proofProfile, inMode, outMode, renderingIntent=..., proofRenderingIntent=..., flags=...
): ...

buildTransformFromOpenProfiles = buildTransform
buildProofTransformFromOpenProfiles = buildProofTransform

def applyTransform(im, transform, inPlace: bool = ...): ...
def createProfile(colorSpace, colorTemp: int = ...): ...
def getProfileName(profile): ...
def getProfileInfo(profile): ...
def getProfileCopyright(profile): ...
def getProfileManufacturer(profile): ...
def getProfileModel(profile): ...
def getProfileDescription(profile): ...
def getDefaultIntent(profile): ...
def isIntentSupported(profile, intent, direction): ...
def versions(): ...



import enum
import sre_compile
import sys
from sre_constants import error as error
from typing import Any, AnyStr, Callable, Iterator, overload

# ----- re variables and constants -----
if sys.version_info >= (3, 7):
    from typing import Match as Match, Pattern as Pattern
else:
    from typing import Match, Pattern

if sys.version_info >= (3, 11):
    __all__ = [
        "match",
        "fullmatch",
        "search",
        "sub",
        "subn",
        "split",
        "findall",
        "finditer",
        "compile",
        "purge",
        "template",
        "escape",
        "error",
        "Pattern",
        "Match",
        "A",
        "I",
        "L",
        "M",
        "S",
        "X",
        "U",
        "ASCII",
        "IGNORECASE",
        "LOCALE",
        "MULTILINE",
        "DOTALL",
        "VERBOSE",
        "UNICODE",
        "RegexFlag",
        "NOFLAG",
    ]
elif sys.version_info >= (3, 8):
    __all__ = [
        "match",
        "fullmatch",
        "search",
        "sub",
        "subn",
        "split",
        "findall",
        "finditer",
        "compile",
        "purge",
        "template",
        "escape",
        "error",
        "Pattern",
        "Match",
        "A",
        "I",
        "L",
        "M",
        "S",
        "X",
        "U",
        "ASCII",
        "IGNORECASE",
        "LOCALE",
        "MULTILINE",
        "DOTALL",
        "VERBOSE",
        "UNICODE",
    ]
else:
    __all__ = [
        "match",
        "fullmatch",
        "search",
        "sub",
        "subn",
        "split",
        "findall",
        "finditer",
        "compile",
        "purge",
        "template",
        "escape",
        "error",
        "A",
        "I",
        "L",
        "M",
        "S",
        "X",
        "U",
        "ASCII",
        "IGNORECASE",
        "LOCALE",
        "MULTILINE",
        "DOTALL",
        "VERBOSE",
        "UNICODE",
    ]

class RegexFlag(enum.IntFlag):
    A = sre_compile.SRE_FLAG_ASCII
    ASCII = A
    DEBUG = sre_compile.SRE_FLAG_DEBUG
    I = sre_compile.SRE_FLAG_IGNORECASE
    IGNORECASE = I
    L = sre_compile.SRE_FLAG_LOCALE
    LOCALE = L
    M = sre_compile.SRE_FLAG_MULTILINE
    MULTILINE = M
    S = sre_compile.SRE_FLAG_DOTALL
    DOTALL = S
    X = sre_compile.SRE_FLAG_VERBOSE
    VERBOSE = X
    U = sre_compile.SRE_FLAG_UNICODE
    UNICODE = U
    T = sre_compile.SRE_FLAG_TEMPLATE
    TEMPLATE = T
    if sys.version_info >= (3, 11):
        NOFLAG: int

A = RegexFlag.A
ASCII = RegexFlag.ASCII
DEBUG = RegexFlag.DEBUG
I = RegexFlag.I
IGNORECASE = RegexFlag.IGNORECASE
L = RegexFlag.L
LOCALE = RegexFlag.LOCALE
M = RegexFlag.M
MULTILINE = RegexFlag.MULTILINE
S = RegexFlag.S
DOTALL = RegexFlag.DOTALL
X = RegexFlag.X
VERBOSE = RegexFlag.VERBOSE
U = RegexFlag.U
UNICODE = RegexFlag.UNICODE
T = RegexFlag.T
TEMPLATE = RegexFlag.TEMPLATE
if sys.version_info >= (3, 11):
    NOFLAG = RegexFlag.NOFLAG
_FlagsType = int | RegexFlag

if sys.version_info < (3, 7):
    # undocumented
    _pattern_type: type

# Type-wise these overloads are unnecessary, they could also be modeled using
# unions in the parameter types. However mypy has a bug regarding TypeVar
# constraints (https://github.com/python/mypy/issues/11880),
# which limits us here because AnyStr is a constrained TypeVar.

@overload
def compile(pattern: AnyStr, flags: _FlagsType = ...) -> Pattern[AnyStr]: ...
@overload
def compile(pattern: Pattern[AnyStr], flags: _FlagsType = ...) -> Pattern[AnyStr]: ...
@overload
def search(pattern: AnyStr, string: AnyStr, flags: _FlagsType = ...) -> Match[AnyStr] | None: ...
@overload
def search(pattern: Pattern[AnyStr], string: AnyStr, flags: _FlagsType = ...) -> Match[AnyStr] | None: ...
@overload
def match(pattern: AnyStr, string: AnyStr, flags: _FlagsType = ...) -> Match[AnyStr] | None: ...
@overload
def match(pattern: Pattern[AnyStr], string: AnyStr, flags: _FlagsType = ...) -> Match[AnyStr] | None: ...
@overload
def fullmatch(pattern: AnyStr, string: AnyStr, flags: _FlagsType = ...) -> Match[AnyStr] | None: ...
@overload
def fullmatch(pattern: Pattern[AnyStr], string: AnyStr, flags: _FlagsType = ...) -> Match[AnyStr] | None: ...
@overload
def split(pattern: AnyStr, string: AnyStr, maxsplit: int = ..., flags: _FlagsType = ...) -> list[AnyStr | Any]: ...
@overload
def split(pattern: Pattern[AnyStr], string: AnyStr, maxsplit: int = ..., flags: _FlagsType = ...) -> list[AnyStr | Any]: ...
@overload
def findall(pattern: AnyStr, string: AnyStr, flags: _FlagsType = ...) -> list[Any]: ...
@overload
def findall(pattern: Pattern[AnyStr], string: AnyStr, flags: _FlagsType = ...) -> list[Any]: ...

# Return an iterator yielding match objects over all non-overlapping matches
# for the RE pattern in string. The string is scanned left-to-right, and
# matches are returned in the order found. Empty matches are included in the
# result unless they touch the beginning of another match.
@overload
def finditer(pattern: AnyStr, string: AnyStr, flags: _FlagsType = ...) -> Iterator[Match[AnyStr]]: ...
@overload
def finditer(pattern: Pattern[AnyStr], string: AnyStr, flags: _FlagsType = ...) -> Iterator[Match[AnyStr]]: ...
@overload
def sub(pattern: AnyStr, repl: AnyStr, string: AnyStr, count: int = ..., flags: _FlagsType = ...) -> AnyStr: ...
@overload
def sub(
    pattern: AnyStr, repl: Callable[[Match[AnyStr]], AnyStr], string: AnyStr, count: int = ..., flags: _FlagsType = ...
) -> AnyStr: ...
@overload
def sub(pattern: Pattern[AnyStr], repl: AnyStr, string: AnyStr, count: int = ..., flags: _FlagsType = ...) -> AnyStr: ...
@overload
def sub(
    pattern: Pattern[AnyStr], repl: Callable[[Match[AnyStr]], AnyStr], string: AnyStr, count: int = ..., flags: _FlagsType = ...
) -> AnyStr: ...
@overload
def subn(pattern: AnyStr, repl: AnyStr, string: AnyStr, count: int = ..., flags: _FlagsType = ...) -> tuple[AnyStr, int]: ...
@overload
def subn(
    pattern: AnyStr, repl: Callable[[Match[AnyStr]], AnyStr], string: AnyStr, count: int = ..., flags: _FlagsType = ...
) -> tuple[AnyStr, int]: ...
@overload
def subn(
    pattern: Pattern[AnyStr], repl: AnyStr, string: AnyStr, count: int = ..., flags: _FlagsType = ...
) -> tuple[AnyStr, int]: ...
@overload
def subn(
    pattern: Pattern[AnyStr], repl: Callable[[Match[AnyStr]], AnyStr], string: AnyStr, count: int = ..., flags: _FlagsType = ...
) -> tuple[AnyStr, int]: ...
def escape(pattern: AnyStr) -> AnyStr: ...
def purge() -> None: ...
def template(pattern: AnyStr | Pattern[AnyStr], flags: _FlagsType = ...) -> Pattern[AnyStr]: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class CellSmartTagPr(Serialisable):
    tagname: str
    key: Any
    val: Any
    def __init__(self, key: Any | None = ..., val: Any | None = ...) -> None: ...

class CellSmartTag(Serialisable):
    tagname: str
    cellSmartTagPr: Any
    type: Any
    deleted: Any
    xmlBased: Any
    __elements__: Any
    def __init__(self, cellSmartTagPr=..., type: Any | None = ..., deleted: bool = ..., xmlBased: bool = ...) -> None: ...

class CellSmartTags(Serialisable):
    tagname: str
    cellSmartTag: Any
    r: Any
    __elements__: Any
    def __init__(self, cellSmartTag=..., r: Any | None = ...) -> None: ...

class SmartTags(Serialisable):
    tagname: str
    cellSmartTags: Any
    __elements__: Any
    def __init__(self, cellSmartTags=...) -> None: ...



"""Base class for visitors."""

import re
from typing import Any

from pytype.pytd import pytd
from pytype.typegraph import cfg_utils

# A convenient value for unchecked_node_classnames if a visitor wants to
# use unchecked nodes everywhere.
ALL_NODE_NAMES = type(
    "contains_everything",
    (),
    {"__contains__": lambda *args: True})()


class _NodeClassInfo:
  """Representation of a node class in the graph."""

  def __init__(self, cls):
    self.cls = cls  # The class object.
    self.name = cls.__name__
    # The set of NodeClassInfo objects that may appear below this particular
    # type of node. Initially empty, filled in by examining child fields.
    self.outgoing = set()


def _FindNodeClasses():
  """Yields _NodeClassInfo objects for each node found in pytd."""
  for name in dir(pytd):
    value = getattr(pytd, name)
    if (isinstance(value, type) and
        issubclass(value, pytd.Node) and
        value is not pytd.Node):
      yield _NodeClassInfo(value)


_IGNORED_TYPES = frozenset([str, bool, int, type(None), Any])
_ancestor_map = None  # Memoized ancestors map.


def _GetChildTypes(node_classes, cls: Any):
  """Get all the types that can be in a node's subtree."""

  types = set()

  def AddType(t: Any):
    if hasattr(t, "__args__"):
      # Tuple[...] and Union[...] store their contained types in __args__
      for x in t.__args__:
        if x is not Ellipsis:
          AddType(x)
      return
    if hasattr(t, "__forward_arg__"):
      # __forward_arg__ is the runtime representation of late annotations
      t = t.__forward_arg__
    if isinstance(t, str) and t in node_classes:
      types.add(node_classes[t].cls)
    else:
      types.add(t)

  # Directly accessing __attrs_attrs__ is faster than calling attr.fields.
  for field in cls.__attrs_attrs__:  # pytype: disable=attribute-error
    AddType(field.type)

  # Verify that all late types have been converted.
  for x in types:
    assert isinstance(x, type) or x == Any

  return types


def _GetAncestorMap():
  """Return a map of node class names to a set of ancestor class names."""

  global _ancestor_map
  if _ancestor_map is None:
    # Map from name to _NodeClassInfo.
    node_classes = {i.name: i for i in _FindNodeClasses()}

    # Update _NodeClassInfo.outgoing based on children.
    for info in node_classes.values():
      for allowed in _GetChildTypes(node_classes, info.cls):
        if allowed in _IGNORED_TYPES:
          pass
        elif allowed.__module__ == "pytype.pytd.pytd":
          # All subclasses of the type are allowed.
          info.outgoing.update(
              [i for i in node_classes.values() if issubclass(i.cls, allowed)])
        else:
          # This means we have a child type that is unknown. If it is a node
          # then make sure _FindNodeClasses() can discover it. If it is not a
          # node, then add the typename to _IGNORED_TYPES.
          raise AssertionError(f"Unknown child type: {allowed}")

    predecessors = cfg_utils.compute_predecessors(node_classes.values())
    # Convert predecessors keys and values to use names instead of info objects.
    get_names = lambda v: {n.name for n in v}
    _ancestor_map = {k.name: get_names(v) for k, v in predecessors.items()}
  return _ancestor_map


class Visitor:
  """Base class for visitors.

  Each class inheriting from visitor SHOULD have a fixed set of methods,
  otherwise it might break the caching in this class.

  Attributes:
    visits_all_node_types: Whether the visitor can visit every node type.
    unchecked_node_names: Contains the names of node classes that are unchecked
      when constructing a new node from visited children.  This is useful
      if a visitor returns data in part or all of its walk that would violate
      node preconditions.
    enter_functions: A dictionary mapping node class names to the
      corresponding Enter functions.
    visit_functions: A dictionary mapping node class names to the
      corresponding Visit functions.
    leave_functions: A dictionary mapping node class names to the
      corresponding Leave functions.
    visit_class_names: A set of node class names that must be visited.  This is
      constructed based on the enter/visit/leave functions and precondition
      data about legal ASTs.  As an optimization, the visitor will only visit
      nodes under which some actionable node can appear.
  """
  # The old_node attribute contains a copy of the node before its children were
  # visited. It has the same type as the node currently being visited.
  old_node: Any

  visits_all_node_types = False
  unchecked_node_names = set()

  _visitor_functions_cache = {}

  def __init__(self):
    cls = self.__class__

    # The set of method names for each visitor implementation is assumed to
    # be fixed. Therefore this introspection can be cached.
    if cls in Visitor._visitor_functions_cache:
      enter_fns, visit_fns, leave_fns, visit_class_names = (
          Visitor._visitor_functions_cache[cls])
    else:
      enter_fns = {}
      enter_prefix = "Enter"
      enter_len = len(enter_prefix)

      visit_fns = {}
      visit_prefix = "Visit"
      visit_len = len(visit_prefix)

      leave_fns = {}
      leave_prefix = "Leave"
      leave_len = len(leave_prefix)

      for attrib in dir(cls):
        if attrib.startswith(enter_prefix):
          enter_fns[attrib[enter_len:]] = getattr(cls, attrib)
        elif attrib.startswith(visit_prefix):
          visit_fns[attrib[visit_len:]] = getattr(cls, attrib)
        elif attrib.startswith(leave_prefix):
          leave_fns[attrib[leave_len:]] = getattr(cls, attrib)

      ancestors = _GetAncestorMap()
      visit_class_names = set()
      # A custom Enter/Visit/Leave requires visiting all types of nodes.
      visit_all = (cls.Enter != Visitor.Enter or
                   cls.Visit != Visitor.Visit or
                   cls.Leave != Visitor.Leave)
      for node in set(enter_fns) | set(visit_fns) | set(leave_fns):
        if node in ancestors:
          visit_class_names.update(ancestors[node])
        elif node:
          # Visiting an unknown non-empty node means the visitor has defined
          # behavior on nodes that are unknown to the ancestors list.
          if node == "StrictType":
            # This special case is here because pytd.type_match defines an extra
            # StrictType node, and pytd.printer.PrintVisitor has a visitor to
            # handle it.
            visit_all = True
          elif (cls.__module__ == "__main__" or
                re.fullmatch(r".*(_test|test_[^\.]+)", cls.__module__)):
            # We are running test code or something else that is defining its
            # own pytd nodes directly in a top-level python file.
            visit_all = True
          else:
            raise AssertionError(f"Unknown node type: {node} {cls!r}")
      if visit_all:
        visit_class_names = ALL_NODE_NAMES
      Visitor._visitor_functions_cache[cls] = (
          enter_fns, visit_fns, leave_fns, visit_class_names)

    self.enter_functions = enter_fns
    self.visit_functions = visit_fns
    self.leave_functions = leave_fns
    self.visit_class_names = visit_class_names

  def Enter(self, node, *args, **kwargs):
    return self.enter_functions[node.__class__.__name__](
        self, node, *args, **kwargs)

  def Visit(self, node, *args, **kwargs):
    return self.visit_functions[node.__class__.__name__](
        self, node, *args, **kwargs)

  def Leave(self, node, *args, **kwargs):
    self.leave_functions[node.__class__.__name__](self, node, *args, **kwargs)



import sys
from _typeshed import Self
from typing import Iterable, TextIO

if sys.version_info >= (3, 8):
    __all__ = ["shlex", "split", "quote", "join"]
else:
    __all__ = ["shlex", "split", "quote"]

def split(s: str, comments: bool = ..., posix: bool = ...) -> list[str]: ...

if sys.version_info >= (3, 8):
    def join(split_command: Iterable[str]) -> str: ...

def quote(s: str) -> str: ...

class shlex(Iterable[str]):
    commenters: str
    wordchars: str
    whitespace: str
    escape: str
    quotes: str
    escapedquotes: str
    whitespace_split: bool
    infile: str | None
    instream: TextIO
    source: str
    debug: int
    lineno: int
    token: str
    eof: str
    if sys.version_info >= (3, 7):
        @property
        def punctuation_chars(self) -> str: ...
    else:
        punctuation_chars: str
    def __init__(
        self,
        instream: str | TextIO | None = ...,
        infile: str | None = ...,
        posix: bool = ...,
        punctuation_chars: bool | str = ...,
    ) -> None: ...
    def get_token(self) -> str: ...
    def push_token(self, tok: str) -> None: ...
    def read_token(self) -> str: ...
    def sourcehook(self, newfile: str) -> tuple[str, TextIO]: ...
    def push_source(self, newstream: str | TextIO, newfile: str | None = ...) -> None: ...
    def pop_source(self) -> None: ...
    def error_leader(self, infile: str | None = ..., lineno: int | None = ...) -> None: ...
    def __iter__(self: Self) -> Self: ...
    def __next__(self) -> str: ...



from typing import Any

from braintree.account_updater_daily_report import AccountUpdaterDailyReport as AccountUpdaterDailyReport
from braintree.configuration import Configuration as Configuration
from braintree.connected_merchant_paypal_status_changed import (
    ConnectedMerchantPayPalStatusChanged as ConnectedMerchantPayPalStatusChanged,
)
from braintree.connected_merchant_status_transitioned import (
    ConnectedMerchantStatusTransitioned as ConnectedMerchantStatusTransitioned,
)
from braintree.disbursement import Disbursement as Disbursement
from braintree.dispute import Dispute as Dispute
from braintree.error_result import ErrorResult as ErrorResult
from braintree.granted_payment_instrument_update import GrantedPaymentInstrumentUpdate as GrantedPaymentInstrumentUpdate
from braintree.local_payment_completed import LocalPaymentCompleted as LocalPaymentCompleted
from braintree.local_payment_reversed import LocalPaymentReversed as LocalPaymentReversed
from braintree.merchant_account import MerchantAccount as MerchantAccount
from braintree.oauth_access_revocation import OAuthAccessRevocation as OAuthAccessRevocation
from braintree.partner_merchant import PartnerMerchant as PartnerMerchant
from braintree.resource import Resource as Resource
from braintree.revoked_payment_method_metadata import RevokedPaymentMethodMetadata as RevokedPaymentMethodMetadata
from braintree.subscription import Subscription as Subscription
from braintree.transaction import Transaction as Transaction
from braintree.validation_error_collection import ValidationErrorCollection as ValidationErrorCollection

class WebhookNotification(Resource):
    class Kind:
        AccountUpdaterDailyReport: str
        Check: str
        ConnectedMerchantPayPalStatusChanged: str
        ConnectedMerchantStatusTransitioned: str
        Disbursement: str
        DisbursementException: str
        DisputeAccepted: str
        DisputeDisputed: str
        DisputeExpired: str
        DisputeLost: str
        DisputeOpened: str
        DisputeWon: str
        GrantedPaymentMethodRevoked: str
        GrantorUpdatedGrantedPaymentMethod: str
        LocalPaymentCompleted: str
        LocalPaymentReversed: str
        OAuthAccessRevoked: str
        PartnerMerchantConnected: str
        PartnerMerchantDeclined: str
        PartnerMerchantDisconnected: str
        PaymentMethodRevokedByCustomer: str
        RecipientUpdatedGrantedPaymentMethod: str
        SubMerchantAccountApproved: str
        SubMerchantAccountDeclined: str
        SubscriptionCanceled: str
        SubscriptionChargedSuccessfully: str
        SubscriptionChargedUnsuccessfully: str
        SubscriptionExpired: str
        SubscriptionTrialEnded: str
        SubscriptionWentActive: str
        SubscriptionWentPastDue: str
        TransactionDisbursed: str
        TransactionSettled: str
        TransactionSettlementDeclined: str
    @staticmethod
    def parse(signature, payload): ...
    @staticmethod
    def verify(challenge): ...
    source_merchant_id: Any
    subscription: Any
    merchant_account: Any
    transaction: Any
    connected_merchant_status_transitioned: Any
    connected_merchant_paypal_status_changed: Any
    partner_merchant: Any
    oauth_access_revocation: Any
    disbursement: Any
    dispute: Any
    account_updater_daily_report: Any
    granted_payment_instrument_update: Any
    revoked_payment_method_metadata: Any
    local_payment_completed: Any
    local_payment_reversed: Any
    errors: Any
    message: Any
    def __init__(self, gateway, attributes) -> None: ...



import importlib.abc
import types
import zipimport
from _typeshed import Self
from abc import ABCMeta
from typing import IO, Any, Callable, Generator, Iterable, Sequence, TypeVar, overload

LegacyVersion = Any  # from packaging.version
Version = Any  # from packaging.version

_T = TypeVar("_T")
_NestedStr = str | Iterable[str | Iterable[Any]]
_InstallerType = Callable[[Requirement], Distribution | None]
_EPDistType = Distribution | Requirement | str
_MetadataType = IResourceProvider | None
_PkgReqType = str | Requirement
_DistFinderType = Callable[[_Importer, str, bool], Generator[Distribution, None, None]]
_NSHandlerType = Callable[[_Importer, str, str, types.ModuleType], str]

def declare_namespace(name: str) -> None: ...
def fixup_namespace_packages(path_item: str) -> None: ...

class WorkingSet:
    entries: list[str]
    def __init__(self, entries: Iterable[str] | None = ...) -> None: ...
    def require(self, *requirements: _NestedStr) -> Sequence[Distribution]: ...
    def run_script(self, requires: str, script_name: str) -> None: ...
    def iter_entry_points(self, group: str, name: str | None = ...) -> Generator[EntryPoint, None, None]: ...
    def add_entry(self, entry: str) -> None: ...
    def __contains__(self, dist: Distribution) -> bool: ...
    def __iter__(self) -> Generator[Distribution, None, None]: ...
    def find(self, req: Requirement) -> Distribution | None: ...
    def resolve(
        self, requirements: Iterable[Requirement], env: Environment | None = ..., installer: _InstallerType | None = ...
    ) -> list[Distribution]: ...
    def add(self, dist: Distribution, entry: str | None = ..., insert: bool = ..., replace: bool = ...) -> None: ...
    def subscribe(self, callback: Callable[[Distribution], None]) -> None: ...
    def find_plugins(
        self, plugin_env: Environment, full_env: Environment | None = ..., fallback: bool = ...
    ) -> tuple[list[Distribution], dict[Distribution, Exception]]: ...

working_set: WorkingSet = ...

require = working_set.require
run_script = working_set.run_script
iter_entry_points = working_set.iter_entry_points
add_activation_listener = working_set.subscribe

class Environment:
    def __init__(self, search_path: Sequence[str] | None = ..., platform: str | None = ..., python: str | None = ...) -> None: ...
    def __getitem__(self, project_name: str) -> list[Distribution]: ...
    def __iter__(self) -> Generator[str, None, None]: ...
    def add(self, dist: Distribution) -> None: ...
    def remove(self, dist: Distribution) -> None: ...
    def can_add(self, dist: Distribution) -> bool: ...
    def __add__(self, other: Distribution | Environment) -> Environment: ...
    def __iadd__(self: Self, other: Distribution | Environment) -> Self: ...
    @overload
    def best_match(self, req: Requirement, working_set: WorkingSet) -> Distribution: ...
    @overload
    def best_match(self, req: Requirement, working_set: WorkingSet, installer: Callable[[Requirement], _T] = ...) -> _T: ...
    @overload
    def obtain(self, requirement: Requirement) -> None: ...
    @overload
    def obtain(self, requirement: Requirement, installer: Callable[[Requirement], _T] = ...) -> _T: ...
    def scan(self, search_path: Sequence[str] | None = ...) -> None: ...

def parse_requirements(strs: str | Iterable[str]) -> Generator[Requirement, None, None]: ...

class Requirement:
    unsafe_name: str
    project_name: str
    key: str
    extras: tuple[str, ...]
    specs: list[tuple[str, str]]
    # TODO: change this to packaging.markers.Marker | None once we can import
    #       packaging.markers
    marker: Any | None
    @staticmethod
    def parse(s: str | Iterable[str]) -> Requirement: ...
    def __contains__(self, item: Distribution | str | tuple[str, ...]) -> bool: ...
    def __eq__(self, other_requirement: object) -> bool: ...

def load_entry_point(dist: _EPDistType, group: str, name: str) -> Any: ...
def get_entry_info(dist: _EPDistType, group: str, name: str) -> EntryPoint | None: ...
@overload
def get_entry_map(dist: _EPDistType) -> dict[str, dict[str, EntryPoint]]: ...
@overload
def get_entry_map(dist: _EPDistType, group: str) -> dict[str, EntryPoint]: ...

class EntryPoint:
    name: str
    module_name: str
    attrs: tuple[str, ...]
    extras: tuple[str, ...]
    dist: Distribution | None
    def __init__(
        self,
        name: str,
        module_name: str,
        attrs: tuple[str, ...] = ...,
        extras: tuple[str, ...] = ...,
        dist: Distribution | None = ...,
    ) -> None: ...
    @classmethod
    def parse(cls, src: str, dist: Distribution | None = ...) -> EntryPoint: ...
    @classmethod
    def parse_group(cls, group: str, lines: str | Sequence[str], dist: Distribution | None = ...) -> dict[str, EntryPoint]: ...
    @classmethod
    def parse_map(
        cls, data: dict[str, str | Sequence[str]] | str | Sequence[str], dist: Distribution | None = ...
    ) -> dict[str, EntryPoint]: ...
    def load(self, require: bool = ..., env: Environment | None = ..., installer: _InstallerType | None = ...) -> Any: ...
    def require(self, env: Environment | None = ..., installer: _InstallerType | None = ...) -> None: ...
    def resolve(self) -> Any: ...

def find_distributions(path_item: str, only: bool = ...) -> Generator[Distribution, None, None]: ...
def get_distribution(dist: Requirement | str | Distribution) -> Distribution: ...

class Distribution(IResourceProvider, IMetadataProvider):
    PKG_INFO: str
    location: str
    project_name: str
    @property
    def key(self) -> str: ...
    @property
    def extras(self) -> list[str]: ...
    @property
    def version(self) -> str: ...
    @property
    def parsed_version(self) -> tuple[str, ...]: ...
    py_version: str
    platform: str | None
    precedence: int
    def __init__(
        self,
        location: str | None = ...,
        metadata: _MetadataType = ...,
        project_name: str | None = ...,
        version: str | None = ...,
        py_version: str = ...,
        platform: str | None = ...,
        precedence: int = ...,
    ) -> None: ...
    @classmethod
    def from_location(
        cls, location: str, basename: str, metadata: _MetadataType = ..., **kw: str | None | int
    ) -> Distribution: ...
    @classmethod
    def from_filename(cls, filename: str, metadata: _MetadataType = ..., **kw: str | None | int) -> Distribution: ...
    def activate(self, path: list[str] | None = ...) -> None: ...
    def as_requirement(self) -> Requirement: ...
    def requires(self, extras: tuple[str, ...] = ...) -> list[Requirement]: ...
    def clone(self, **kw: str | int | None) -> Requirement: ...
    def egg_name(self) -> str: ...
    def __cmp__(self, other: Any) -> bool: ...
    def get_entry_info(self, group: str, name: str) -> EntryPoint | None: ...
    @overload
    def get_entry_map(self) -> dict[str, dict[str, EntryPoint]]: ...
    @overload
    def get_entry_map(self, group: str) -> dict[str, EntryPoint]: ...
    def load_entry_point(self, group: str, name: str) -> Any: ...

EGG_DIST: int
BINARY_DIST: int
SOURCE_DIST: int
CHECKOUT_DIST: int
DEVELOP_DIST: int

def resource_exists(package_or_requirement: _PkgReqType, resource_name: str) -> bool: ...
def resource_stream(package_or_requirement: _PkgReqType, resource_name: str) -> IO[bytes]: ...
def resource_string(package_or_requirement: _PkgReqType, resource_name: str) -> bytes: ...
def resource_isdir(package_or_requirement: _PkgReqType, resource_name: str) -> bool: ...
def resource_listdir(package_or_requirement: _PkgReqType, resource_name: str) -> list[str]: ...
def resource_filename(package_or_requirement: _PkgReqType, resource_name: str) -> str: ...
def set_extraction_path(path: str) -> None: ...
def cleanup_resources(force: bool = ...) -> list[str]: ...

class IResourceManager:
    def resource_exists(self, package_or_requirement: _PkgReqType, resource_name: str) -> bool: ...
    def resource_stream(self, package_or_requirement: _PkgReqType, resource_name: str) -> IO[bytes]: ...
    def resource_string(self, package_or_requirement: _PkgReqType, resource_name: str) -> bytes: ...
    def resource_isdir(self, package_or_requirement: _PkgReqType, resource_name: str) -> bool: ...
    def resource_listdir(self, package_or_requirement: _PkgReqType, resource_name: str) -> list[str]: ...
    def resource_filename(self, package_or_requirement: _PkgReqType, resource_name: str) -> str: ...
    def set_extraction_path(self, path: str) -> None: ...
    def cleanup_resources(self, force: bool = ...) -> list[str]: ...
    def get_cache_path(self, archive_name: str, names: Iterable[str] = ...) -> str: ...
    def extraction_error(self) -> None: ...
    def postprocess(self, tempname: str, filename: str) -> None: ...

@overload
def get_provider(package_or_requirement: str) -> IResourceProvider: ...
@overload
def get_provider(package_or_requirement: Requirement) -> Distribution: ...

class IMetadataProvider:
    def has_metadata(self, name: str) -> bool: ...
    def metadata_isdir(self, name: str) -> bool: ...
    def metadata_listdir(self, name: str) -> list[str]: ...
    def get_metadata(self, name: str) -> str: ...
    def get_metadata_lines(self, name: str) -> Generator[str, None, None]: ...
    def run_script(self, script_name: str, namespace: dict[str, Any]) -> None: ...

class ResolutionError(Exception): ...

class DistributionNotFound(ResolutionError):
    @property
    def req(self) -> Requirement: ...
    @property
    def requirers(self) -> set[str]: ...
    @property
    def requirers_str(self) -> str: ...
    def report(self) -> str: ...

class VersionConflict(ResolutionError):
    @property
    def dist(self) -> Any: ...
    @property
    def req(self) -> Any: ...
    def report(self) -> str: ...
    def with_context(self, required_by: set[Distribution | str]) -> VersionConflict: ...

class ContextualVersionConflict(VersionConflict):
    @property
    def required_by(self) -> set[Distribution | str]: ...

class UnknownExtra(ResolutionError): ...

class ExtractionError(Exception):
    manager: IResourceManager
    cache_path: str
    original_error: Exception

class _Importer(importlib.abc.MetaPathFinder, importlib.abc.InspectLoader, metaclass=ABCMeta): ...

def register_finder(importer_type: type, distribution_finder: _DistFinderType) -> None: ...
def register_loader_type(loader_type: type, provider_factory: Callable[[types.ModuleType], IResourceProvider]) -> None: ...
def register_namespace_handler(importer_type: type, namespace_handler: _NSHandlerType) -> None: ...

class IResourceProvider(IMetadataProvider): ...
class NullProvider: ...
class EggProvider(NullProvider): ...
class DefaultProvider(EggProvider): ...

class PathMetadata(DefaultProvider, IResourceProvider):
    def __init__(self, path: str, egg_info: str) -> None: ...

class ZipProvider(EggProvider): ...

class EggMetadata(ZipProvider, IResourceProvider):
    def __init__(self, zipimporter: zipimport.zipimporter) -> None: ...

class EmptyProvider(NullProvider): ...

empty_provider: EmptyProvider

class FileMetadata(EmptyProvider, IResourceProvider):
    def __init__(self, path_to_pkg_info: str) -> None: ...

def parse_version(v: str) -> Version | LegacyVersion: ...
def yield_lines(strs: _NestedStr) -> Generator[str, None, None]: ...
def split_sections(strs: _NestedStr) -> Generator[tuple[str | None, str], None, None]: ...
def safe_name(name: str) -> str: ...
def safe_version(version: str) -> str: ...
def safe_extra(extra: str) -> str: ...
def to_filename(name_or_version: str) -> str: ...
def get_build_platform() -> str: ...
def get_platform() -> str: ...
def get_supported_platform() -> str: ...
def compatible_platforms(provided: str | None, required: str | None) -> bool: ...
def get_default_cache() -> str: ...
def get_importer(path_item: str) -> _Importer: ...
def ensure_directory(path: str) -> None: ...
def normalize_path(filename: str) -> str: ...



from typing import Iterable, Mapping, NoReturn

from paramiko.channel import Channel, ChannelFile, ChannelStderrFile, ChannelStdinFile
from paramiko.hostkeys import HostKeys
from paramiko.pkey import PKey
from paramiko.sftp_client import SFTPClient
from paramiko.transport import Transport
from paramiko.util import ClosingContextManager

from .transport import _SocketLike

class SSHClient(ClosingContextManager):
    def __init__(self) -> None: ...
    def load_system_host_keys(self, filename: str | None = ...) -> None: ...
    def load_host_keys(self, filename: str) -> None: ...
    def save_host_keys(self, filename: str) -> None: ...
    def get_host_keys(self) -> HostKeys: ...
    def set_log_channel(self, name: str) -> None: ...
    def set_missing_host_key_policy(self, policy: type[MissingHostKeyPolicy] | MissingHostKeyPolicy) -> None: ...
    def connect(
        self,
        hostname: str,
        port: int = ...,
        username: str | None = ...,
        password: str | None = ...,
        pkey: PKey | None = ...,
        key_filename: str | None = ...,
        timeout: float | None = ...,
        allow_agent: bool = ...,
        look_for_keys: bool = ...,
        compress: bool = ...,
        sock: _SocketLike | None = ...,
        gss_auth: bool = ...,
        gss_kex: bool = ...,
        gss_deleg_creds: bool = ...,
        gss_host: str | None = ...,
        banner_timeout: float | None = ...,
        auth_timeout: float | None = ...,
        gss_trust_dns: bool = ...,
        passphrase: str | None = ...,
        disabled_algorithms: dict[str, Iterable[str]] | None = ...,
    ) -> None: ...
    def close(self) -> None: ...
    def exec_command(
        self,
        command: str,
        bufsize: int = ...,
        timeout: float | None = ...,
        get_pty: bool = ...,
        environment: dict[str, str] | None = ...,
    ) -> tuple[ChannelStdinFile, ChannelFile, ChannelStderrFile]: ...
    def invoke_shell(
        self,
        term: str = ...,
        width: int = ...,
        height: int = ...,
        width_pixels: int = ...,
        height_pixels: int = ...,
        environment: Mapping[str, str] | None = ...,
    ) -> Channel: ...
    def open_sftp(self) -> SFTPClient: ...
    def get_transport(self) -> Transport | None: ...

class MissingHostKeyPolicy:
    def missing_host_key(self, client: SSHClient, hostname: str, key: PKey) -> None: ...

class AutoAddPolicy(MissingHostKeyPolicy):
    def missing_host_key(self, client: SSHClient, hostname: str, key: PKey) -> None: ...

class RejectPolicy(MissingHostKeyPolicy):
    def missing_host_key(self, client: SSHClient, hostname: str, key: PKey) -> NoReturn: ...

class WarningPolicy(MissingHostKeyPolicy):
    def missing_host_key(self, client: SSHClient, hostname: str, key: PKey) -> None: ...



import sys

# As at runtime, this depends on all submodules defining __all__ accurately.
from .base_events import *
from .coroutines import *
from .events import *
from .futures import *
from .locks import *
from .protocols import *
from .queues import *
from .streams import *
from .subprocess import *
from .tasks import *
from .transports import *

if sys.version_info >= (3, 7):
    from .runners import *

if sys.version_info >= (3, 8):
    from .exceptions import *

if sys.version_info >= (3, 9):
    from .threads import *

if sys.version_info >= (3, 11):
    from .taskgroups import *

if sys.platform == "win32":
    from .windows_events import *
else:
    from .unix_events import *



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable
from openpyxl.styles.fonts import Font

class PhoneticProperties(Serialisable):
    tagname: str
    fontId: Any
    type: Any
    alignment: Any
    def __init__(self, fontId: Any | None = ..., type: Any | None = ..., alignment: Any | None = ...) -> None: ...

class PhoneticText(Serialisable):
    tagname: str
    sb: Any
    eb: Any
    t: Any
    text: Any
    def __init__(self, sb: Any | None = ..., eb: Any | None = ..., t: Any | None = ...) -> None: ...

class InlineFont(Font):
    tagname: str
    rFont: Any
    charset: Any
    family: Any
    b: Any
    i: Any
    strike: Any
    outline: Any
    shadow: Any
    condense: Any
    extend: Any
    color: Any
    sz: Any
    u: Any
    vertAlign: Any
    scheme: Any
    __elements__: Any
    def __init__(
        self,
        rFont: Any | None = ...,
        charset: Any | None = ...,
        family: Any | None = ...,
        b: Any | None = ...,
        i: Any | None = ...,
        strike: Any | None = ...,
        outline: Any | None = ...,
        shadow: Any | None = ...,
        condense: Any | None = ...,
        extend: Any | None = ...,
        color: Any | None = ...,
        sz: Any | None = ...,
        u: Any | None = ...,
        vertAlign: Any | None = ...,
        scheme: Any | None = ...,
    ) -> None: ...

class RichText(Serialisable):
    tagname: str
    rPr: Any
    font: Any
    t: Any
    text: Any
    __elements__: Any
    def __init__(self, rPr: Any | None = ..., t: Any | None = ...) -> None: ...

class Text(Serialisable):
    tagname: str
    t: Any
    plain: Any
    r: Any
    formatted: Any
    rPh: Any
    phonetic: Any
    phoneticPr: Any
    PhoneticProperties: Any
    __elements__: Any
    def __init__(self, t: Any | None = ..., r=..., rPh=..., phoneticPr: Any | None = ...) -> None: ...
    @property
    def content(self): ...



from typing import Iterable

def smart_truncate(
    string: str, max_length: int = ..., word_boundary: bool = ..., separator: str = ..., save_order: bool = ...
) -> str: ...
def slugify(
    text: str,
    entities: bool = ...,
    decimal: bool = ...,
    hexadecimal: bool = ...,
    max_length: int = ...,
    word_boundary: bool = ...,
    separator: str = ...,
    save_order: bool = ...,
    stopwords: Iterable[str] = ...,
    regex_pattern: str | None = ...,
    lowercase: bool = ...,
    replacements: Iterable[Iterable[str]] = ...,
) -> str: ...



from distutils.errors import DistutilsError
from typing import Any

class UnpickleableException(Exception):
    @staticmethod
    def dump(type, exc): ...

class ExceptionSaver:
    def __enter__(self): ...
    def __exit__(self, type, exc, tb): ...
    def resume(self) -> None: ...

def run_setup(setup_script, args): ...

class AbstractSandbox:
    def __init__(self) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, exc_type, exc_value, traceback) -> None: ...
    def run(self, func): ...

class DirectorySandbox(AbstractSandbox):
    write_ops: Any
    def __init__(self, sandbox, exceptions=...) -> None: ...
    def tmpnam(self) -> None: ...
    def open(self, file, flags, mode: int = ..., *args, **kw): ...

class SandboxViolation(DistutilsError):
    tmpl: Any



from typing import Any as _Any

# Enable when pyasn1 gets stubs:
# from pyasn1.type.univ import Boolean, Choice, Enumerated, Integer, Null, OctetString, Sequence, SequenceOf, SetOf
Boolean = _Any
Choice = _Any
Enumerated = _Any
Integer = _Any
Null = _Any
OctetString = _Any
Sequence = _Any
SequenceOf = _Any
SetOf = _Any

LDAP_MAX_INT: int
MAXINT: _Any
rangeInt0ToMaxConstraint: _Any
rangeInt1To127Constraint: _Any
size1ToMaxConstraint: _Any
responseValueConstraint: _Any
numericOIDConstraint: _Any
distinguishedNameConstraint: _Any
nameComponentConstraint: _Any
attributeDescriptionConstraint: _Any
uriConstraint: _Any
attributeSelectorConstraint: _Any

class Integer0ToMax(Integer):
    subtypeSpec: _Any

class LDAPString(OctetString):
    encoding: str

class MessageID(Integer0ToMax): ...
class LDAPOID(OctetString): ...
class LDAPDN(LDAPString): ...
class RelativeLDAPDN(LDAPString): ...
class AttributeDescription(LDAPString): ...

class AttributeValue(OctetString):
    encoding: str

class AssertionValue(OctetString):
    encoding: str

class AttributeValueAssertion(Sequence):
    componentType: _Any

class MatchingRuleId(LDAPString): ...

class Vals(SetOf):
    componentType: _Any

class ValsAtLeast1(SetOf):
    componentType: _Any
    subtypeSpec: _Any

class PartialAttribute(Sequence):
    componentType: _Any

class Attribute(Sequence):
    componentType: _Any

class AttributeList(SequenceOf):
    componentType: _Any

class Simple(OctetString):
    tagSet: _Any
    encoding: str

class Credentials(OctetString):
    encoding: str

class SaslCredentials(Sequence):
    tagSet: _Any
    componentType: _Any

class SicilyPackageDiscovery(OctetString):
    tagSet: _Any
    encoding: str

class SicilyNegotiate(OctetString):
    tagSet: _Any
    encoding: str

class SicilyResponse(OctetString):
    tagSet: _Any
    encoding: str

class AuthenticationChoice(Choice):
    componentType: _Any

class Version(Integer):
    subtypeSpec: _Any

class ResultCode(Enumerated):
    namedValues: _Any
    subTypeSpec: _Any

class URI(LDAPString): ...

class Referral(SequenceOf):
    tagSet: _Any
    componentType: _Any

class ServerSaslCreds(OctetString):
    tagSet: _Any
    encoding: str

class LDAPResult(Sequence):
    componentType: _Any

class Criticality(Boolean):
    defaultValue: bool

class ControlValue(OctetString):
    encoding: str

class Control(Sequence):
    componentType: _Any

class Controls(SequenceOf):
    tagSet: _Any
    componentType: _Any

class Scope(Enumerated):
    namedValues: _Any

class DerefAliases(Enumerated):
    namedValues: _Any

class TypesOnly(Boolean): ...
class Selector(LDAPString): ...

class AttributeSelection(SequenceOf):
    componentType: _Any

class MatchingRule(MatchingRuleId):
    tagSet: _Any

class Type(AttributeDescription):
    tagSet: _Any

class MatchValue(AssertionValue):
    tagSet: _Any

class DnAttributes(Boolean):
    tagSet: _Any
    defaultValue: _Any

class MatchingRuleAssertion(Sequence):
    componentType: _Any

class Initial(AssertionValue):
    tagSet: _Any

class Any(AssertionValue):
    tagSet: _Any

class Final(AssertionValue):
    tagSet: _Any

class Substring(Choice):
    componentType: _Any

class Substrings(SequenceOf):
    subtypeSpec: _Any
    componentType: _Any

class SubstringFilter(Sequence):
    tagSet: _Any
    componentType: _Any

class And(SetOf):
    tagSet: _Any
    subtypeSpec: _Any

class Or(SetOf):
    tagSet: _Any
    subtypeSpec: _Any

class Not(Choice): ...

class EqualityMatch(AttributeValueAssertion):
    tagSet: _Any

class GreaterOrEqual(AttributeValueAssertion):
    tagSet: _Any

class LessOrEqual(AttributeValueAssertion):
    tagSet: _Any

class Present(AttributeDescription):
    tagSet: _Any

class ApproxMatch(AttributeValueAssertion):
    tagSet: _Any

class ExtensibleMatch(MatchingRuleAssertion):
    tagSet: _Any

class Filter(Choice):
    componentType: _Any

class PartialAttributeList(SequenceOf):
    componentType: _Any

class Operation(Enumerated):
    namedValues: _Any

class Change(Sequence):
    componentType: _Any

class Changes(SequenceOf):
    componentType: _Any

class DeleteOldRDN(Boolean): ...

class NewSuperior(LDAPDN):
    tagSet: _Any

class RequestName(LDAPOID):
    tagSet: _Any

class RequestValue(OctetString):
    tagSet: _Any
    encoding: str

class ResponseName(LDAPOID):
    tagSet: _Any

class ResponseValue(OctetString):
    tagSet: _Any
    encoding: str

class IntermediateResponseName(LDAPOID):
    tagSet: _Any

class IntermediateResponseValue(OctetString):
    tagSet: _Any
    encoding: str

class BindRequest(Sequence):
    tagSet: _Any
    componentType: _Any

class BindResponse(Sequence):
    tagSet: _Any
    componentType: _Any

class UnbindRequest(Null):
    tagSet: _Any

class SearchRequest(Sequence):
    tagSet: _Any
    componentType: _Any

class SearchResultReference(SequenceOf):
    tagSet: _Any
    subtypeSpec: _Any
    componentType: _Any

class SearchResultEntry(Sequence):
    tagSet: _Any
    componentType: _Any

class SearchResultDone(LDAPResult):
    tagSet: _Any

class ModifyRequest(Sequence):
    tagSet: _Any
    componentType: _Any

class ModifyResponse(LDAPResult):
    tagSet: _Any

class AddRequest(Sequence):
    tagSet: _Any
    componentType: _Any

class AddResponse(LDAPResult):
    tagSet: _Any

class DelRequest(LDAPDN):
    tagSet: _Any

class DelResponse(LDAPResult):
    tagSet: _Any

class ModifyDNRequest(Sequence):
    tagSet: _Any
    componentType: _Any

class ModifyDNResponse(LDAPResult):
    tagSet: _Any

class CompareRequest(Sequence):
    tagSet: _Any
    componentType: _Any

class CompareResponse(LDAPResult):
    tagSet: _Any

class AbandonRequest(MessageID):
    tagSet: _Any

class ExtendedRequest(Sequence):
    tagSet: _Any
    componentType: _Any

class ExtendedResponse(Sequence):
    tagSet: _Any
    componentType: _Any

class IntermediateResponse(Sequence):
    tagSet: _Any
    componentType: _Any

class ProtocolOp(Choice):
    componentType: _Any

class LDAPMessage(Sequence):
    componentType: _Any



from _typeshed import Self
from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, overload

from .base import _Executable
from .cursor import CursorResult
from .interfaces import Connectable, Dialect
from .url import URL

class MockConnection(Connectable):
    def __init__(self, dialect: Dialect, execute) -> None: ...
    @property
    def engine(self: Self) -> Self: ...  # type: ignore[override]
    @property
    def dialect(self) -> Dialect: ...
    @property
    def name(self) -> str: ...
    def schema_for_object(self, obj): ...
    def connect(self, **kwargs): ...
    def execution_options(self, **kw): ...
    def compiler(self, statement, parameters, **kwargs): ...
    def create(self, entity, **kwargs) -> None: ...
    def drop(self, entity, **kwargs) -> None: ...
    @abstractmethod
    @overload
    def execute(self, object_: _Executable, *multiparams: Mapping[str, Any], **params: Any) -> CursorResult: ...
    @abstractmethod
    @overload
    def execute(self, object_: str, *multiparams: Any | tuple[Any, ...] | Mapping[str, Any], **params: Any) -> CursorResult: ...

def create_mock_engine(url: URL | str, executor, **kw) -> MockConnection: ...



"""Support for PEP593 (typing.Annotated) metadata.

PEP593 allows metadata to be arbitrary python values. We convert these to
strings to support the fact that pytd nodes are immutable, and provide
converters back to python values.

There are two separate forms of encoding: Internally, we want to deal with
metadata as dicts with string keys, which we convert to strings to store in a
pytd.Annotated node and back to dicts via `metadata.{to,from}_string`. When
writing out to a pyi/pytd file, we pretty-print some types to support a more
human-readable annotation, via `metadata.to_pytd`, and parse those back to dicts
in pyi/parser.py.

Note that only a subset of values is supported, see pyi/parser.py
"""

import dataclasses
from typing import Any, Dict, List

from pytype.pyi import evaluator


@dataclasses.dataclass
class Call:
  """Convert callables in metadata to/from generic dicts."""

  fn: str
  posargs: List[Any]
  kwargs: Dict[str, Any]

  @classmethod
  def from_metadata(cls, md, posarg_names, kwarg_names):
    fn = md["tag"]
    posargs = [md[k] for k in posarg_names if k in md]
    kwargs = {k: md[k] for k in kwarg_names if k in md}
    return cls(fn, posargs, kwargs)

  @classmethod
  def from_call_dict(cls, md):
    assert md["tag"] == "call"
    fn = md["fn"]
    posargs = md["posargs"] or []
    kwargs = md["kwargs"] or {}
    return cls(fn, posargs, kwargs)

  def to_metadata(self, posarg_names, kwarg_names):
    out = {"tag": self.fn}
    for name, arg in zip(posarg_names, self.posargs):
      out[name] = arg
    for name in kwarg_names:
      if name in self.kwargs:
        out[name] = self.kwargs[name]
    return out

  def to_call_dict(self):
    return {"tag": "call", "fn": self.fn, "posargs": self.posargs,
            "kwargs": self.kwargs}

  def to_pytd(self):
    posargs = ", ".join(map(repr, self.posargs))
    kwargs = ", ".join(f"{k}={v!r}" for k, v in self.kwargs.items())
    if posargs and kwargs:
      return f"{self.fn}({posargs}, {kwargs})"
    else:
      return f"{self.fn}({posargs}{kwargs})"


# Convert some callables to their own specific metadata dicts.
# {fn: (posarg_names, kwarg_names)}
_CALLABLES = {
    "Deprecated": (["reason"], [])
}


def to_string(val: Any):
  return repr(val)


def from_string(val: str):
  return evaluator.eval_string_literal(val)


def call_to_annotation(fn, *, posargs=None, kwargs=None):
  """Convert a function call to a metadata dict string."""
  call = Call(fn, posargs or (), kwargs or {})
  if fn in _CALLABLES:
    posarg_names, kwarg_names = _CALLABLES[fn]
    out = call.to_metadata(posarg_names, kwarg_names)
  else:
    out = call.to_call_dict()
  return to_string(out)


def to_pytd(metadata: Dict[str, Any]):
  """Convert a metadata dict to a pytd string."""
  tag = metadata.get("tag")
  if tag in _CALLABLES:
    posarg_names, kwarg_names = _CALLABLES[tag]
    return Call.from_metadata(metadata, posarg_names, kwarg_names).to_pytd()
  elif tag == "call":
    return Call.from_call_dict(metadata).to_pytd()
  else:
    return to_string(metadata)



from selenium.webdriver.remote.remote_connection import RemoteConnection as RemoteConnection

class FirefoxRemoteConnection(RemoteConnection):
    def __init__(self, remote_server_addr, keep_alive: bool = ...) -> None: ...



from typing import SupportsComplex, SupportsFloat

e: float
pi: float
_C = SupportsFloat | SupportsComplex | complex

def acos(__z: _C) -> complex: ...
def acosh(__z: _C) -> complex: ...
def asin(__z: _C) -> complex: ...
def asinh(__z: _C) -> complex: ...
def atan(__z: _C) -> complex: ...
def atanh(__z: _C) -> complex: ...
def cos(__z: _C) -> complex: ...
def cosh(__z: _C) -> complex: ...
def exp(__z: _C) -> complex: ...
def isinf(__z: _C) -> bool: ...
def isnan(__z: _C) -> bool: ...
def log(__x: _C, __y_obj: _C = ...) -> complex: ...
def log10(__z: _C) -> complex: ...
def phase(__z: _C) -> float: ...
def polar(__z: _C) -> tuple[float, float]: ...
def rect(__r: float, __phi: float) -> complex: ...
def sin(__z: _C) -> complex: ...
def sinh(__z: _C) -> complex: ...
def sqrt(__z: _C) -> complex: ...
def tan(__z: _C) -> complex: ...
def tanh(__z: _C) -> complex: ...



from typing import Any, Iterator, Protocol, TypeVar

_T = TypeVar("_T")

class PackageMetadata(Protocol):
    def __len__(self) -> int: ...
    def __contains__(self, item: str) -> bool: ...
    def __getitem__(self, key: str) -> str: ...
    def __iter__(self) -> Iterator[str]: ...
    def get_all(self, name: str, failobj: _T = ...) -> list[Any] | _T: ...
    @property
    def json(self) -> dict[str, str | list[str]]: ...

class SimplePath(Protocol):
    def joinpath(self) -> SimplePath: ...
    def __div__(self) -> SimplePath: ...
    def parent(self) -> SimplePath: ...
    def read_text(self) -> str: ...



"""Tests for slots.py."""

from pytype.pytd import slots
import unittest


class TestPytd(unittest.TestCase):
  """Test the operator mappings in slots.py."""

  def test_reverse_name_mapping(self):
    for operator in ("add", "and", "div", "divmod", "floordiv",
                     "lshift", "matmul", "mod", "mul", "or",
                     "pow", "rshift", "sub", "truediv", "xor"):
      normal = f"__{operator}__"
      reverse = f"__r{operator}__"
      self.assertEqual(slots.REVERSE_NAME_MAPPING[normal], reverse)

  def test_symbol_mapping(self):
    for operator, symbol in [("__add__", "+"),
                             ("__invert__", "~")]:
      self.assertEqual(slots.SYMBOL_MAPPING[operator], symbol)


if __name__ == "__main__":
  unittest.main()



from .ansi import Back as Back, Cursor as Cursor, Fore as Fore, Style as Style
from .ansitowin32 import AnsiToWin32 as AnsiToWin32
from .initialise import colorama_text as colorama_text, deinit as deinit, init as init, reinit as reinit



from typing import Any, ClassVar, Iterator, Protocol

from jsonschema._format import FormatChecker
from jsonschema._types import TypeChecker
from jsonschema.exceptions import ValidationError
from jsonschema.validators import RefResolver

class Validator(Protocol):
    META_SCHEMA: ClassVar[dict[Any, Any]]
    VALIDATORS: ClassVar[dict[Any, Any]]
    TYPE_CHECKER: ClassVar[TypeChecker]
    schema: dict[Any, Any] | bool
    def __init__(
        self, schema: dict[Any, Any] | bool, resolver: RefResolver | None = ..., format_checker: FormatChecker | None = ...
    ) -> None: ...
    @classmethod
    def check_schema(cls, schema: dict[Any, Any]) -> None: ...
    def is_type(self, instance: Any, type: str) -> bool: ...
    def is_valid(self, instance: dict[Any, Any]) -> bool: ...
    def iter_errors(self, instance: dict[Any, Any]) -> Iterator[ValidationError]: ...
    def validate(self, instance: dict[Any, Any]) -> None: ...
    def evolve(self, **kwargs) -> Validator: ...



from braintree.util.generator import Generator as Generator
from braintree.util.parser import Parser as Parser

class XmlUtil:
    @staticmethod
    def xml_from_dict(dict): ...
    @staticmethod
    def dict_from_xml(xml): ...



from _typeshed import IdentityFunction
from collections.abc import Iterator, Sequence
from contextlib import AbstractContextManager
from typing import Any, Callable, Generic, MutableMapping, TypeVar, overload

__all__ = ("Cache", "FIFOCache", "LFUCache", "LRUCache", "MRUCache", "RRCache", "TLRUCache", "TTLCache", "cached", "cachedmethod")
__version__: str

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_T = TypeVar("_T")

class Cache(MutableMapping[_KT, _VT], Generic[_KT, _VT]):
    def __init__(self, maxsize: float, getsizeof: Callable[[_VT], float] | None = ...) -> None: ...
    def __getitem__(self, key: _KT) -> _VT: ...
    def __setitem__(self, key: _KT, value: _VT) -> None: ...
    def __delitem__(self, key: _KT) -> None: ...
    def __missing__(self, key: _KT) -> _VT: ...
    def __iter__(self) -> Iterator[_KT]: ...
    def __len__(self) -> int: ...
    @overload  # type: ignore[override]
    def pop(self, key: _KT) -> _VT: ...
    @overload
    def pop(self, key: _KT, default: _VT | _T) -> _VT | _T: ...
    def setdefault(self, key: _KT, default: _VT | None = ...) -> _VT: ...
    @property
    def maxsize(self) -> float: ...
    @property
    def currsize(self) -> float: ...
    @staticmethod
    def getsizeof(value: _VT) -> float: ...

class FIFOCache(Cache[_KT, _VT]):
    def __init__(self, maxsize: float, getsizeof: Callable[[_VT], float] | None = ...) -> None: ...

class LFUCache(Cache[_KT, _VT]):
    def __init__(self, maxsize: float, getsizeof: Callable[[_VT], float] | None = ...) -> None: ...

class LRUCache(Cache[_KT, _VT]):
    def __init__(self, maxsize: float, getsizeof: Callable[[_VT], float] | None = ...) -> None: ...

class MRUCache(Cache[_KT, _VT]):
    def __init__(self, maxsize: float, getsizeof: Callable[[_VT], float] | None = ...) -> None: ...

class RRCache(Cache[_KT, _VT]):
    def __init__(
        self, maxsize: float, choice: Callable[[Sequence[_KT]], _KT] | None = ..., getsizeof: Callable[[_VT], float] | None = ...
    ) -> None: ...
    @property
    def choice(self) -> Callable[[Sequence[_KT]], _KT]: ...

class _TimedCache(Cache[_KT, _VT]):
    def __init__(
        self, maxsize: float, timer: Callable[[], float] = ..., getsizeof: Callable[[_VT], float] | None = ...
    ) -> None: ...
    @property
    def currsize(self) -> float: ...

    class _Timer:
        def __init__(self, timer: Callable[[], float]) -> None: ...
        def __call__(self) -> float: ...
        def __enter__(self) -> float: ...
        def __exit__(self, *exc: object) -> None: ...

    @property
    def timer(self) -> _Timer: ...

class TTLCache(_TimedCache[_KT, _VT]):
    def __init__(
        self, maxsize: float, ttl: float, timer: Callable[[], float] = ..., getsizeof: Callable[[_VT], float] | None = ...
    ) -> None: ...
    @property
    def ttl(self) -> float: ...
    def expire(self, time: float | None = ...) -> None: ...

class TLRUCache(_TimedCache[_KT, _VT]):
    def __init__(
        self,
        maxsize: float,
        ttu: Callable[[_KT, _VT, float], float],
        timer: Callable[[], float] = ...,
        getsizeof: Callable[[_VT], float] | None = ...,
    ) -> None: ...
    @property
    def ttu(self) -> Callable[[_KT, _VT, float], float]: ...
    def expire(self, time: float | None = ...) -> None: ...

def cached(
    cache: MutableMapping[_KT, Any] | None, key: Callable[..., _KT] = ..., lock: AbstractContextManager[Any] | None = ...
) -> IdentityFunction: ...
def cachedmethod(
    cache: Callable[[Any], MutableMapping[_KT, Any] | None],
    key: Callable[..., _KT] = ...,
    lock: Callable[[Any], AbstractContextManager[Any]] | None = ...,
) -> IdentityFunction: ...



from threading import Thread
from typing import Any

from ..strategy.base import BaseStrategy

class AsyncStrategy(BaseStrategy):
    class ReceiverSocketThread(Thread):
        connection: Any
        socket_size: Any
        def __init__(self, ldap_connection) -> None: ...
        def run(self) -> None: ...
    sync: bool
    no_real_dsa: bool
    pooled: bool
    can_stream: bool
    receiver: Any
    async_lock: Any
    event_lock: Any
    def __init__(self, ldap_connection) -> None: ...
    def open(self, reset_usage: bool = ..., read_server_info: bool = ...) -> None: ...
    def close(self) -> None: ...
    def set_event_for_message(self, message_id) -> None: ...
    def post_send_search(self, message_id): ...
    def post_send_single_response(self, message_id): ...
    def receiving(self) -> None: ...
    def get_stream(self) -> None: ...
    def set_stream(self, value) -> None: ...



from typing import Any

from .base import BaseEndpoint as BaseEndpoint

log: Any

class ResourceEndpoint(BaseEndpoint):
    def __init__(self, default_token, token_types) -> None: ...
    @property
    def default_token(self): ...
    @property
    def default_token_type_handler(self): ...
    @property
    def tokens(self): ...
    def verify_request(
        self, uri, http_method: str = ..., body: Any | None = ..., headers: Any | None = ..., scopes: Any | None = ...
    ): ...
    def find_token_type(self, request): ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class MeasureDimensionMap(Serialisable):
    tagname: str
    measureGroup: Any
    dimension: Any
    def __init__(self, measureGroup: Any | None = ..., dimension: Any | None = ...) -> None: ...

class MeasureGroup(Serialisable):
    tagname: str
    name: Any
    caption: Any
    def __init__(self, name: Any | None = ..., caption: Any | None = ...) -> None: ...

class PivotDimension(Serialisable):
    tagname: str
    measure: Any
    name: Any
    uniqueName: Any
    caption: Any
    def __init__(
        self, measure: Any | None = ..., name: Any | None = ..., uniqueName: Any | None = ..., caption: Any | None = ...
    ) -> None: ...

class CalculatedMember(Serialisable):
    tagname: str
    name: Any
    mdx: Any
    memberName: Any
    hierarchy: Any
    parent: Any
    solveOrder: Any
    set: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        name: Any | None = ...,
        mdx: Any | None = ...,
        memberName: Any | None = ...,
        hierarchy: Any | None = ...,
        parent: Any | None = ...,
        solveOrder: Any | None = ...,
        set: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class CalculatedItem(Serialisable):
    tagname: str
    field: Any
    formula: Any
    pivotArea: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self, field: Any | None = ..., formula: Any | None = ..., pivotArea: Any | None = ..., extLst: Any | None = ...
    ) -> None: ...

class ServerFormat(Serialisable):
    tagname: str
    culture: Any
    format: Any
    def __init__(self, culture: Any | None = ..., format: Any | None = ...) -> None: ...

class ServerFormatList(Serialisable):
    tagname: str
    serverFormat: Any
    __elements__: Any
    __attrs__: Any
    def __init__(self, count: Any | None = ..., serverFormat: Any | None = ...) -> None: ...
    @property
    def count(self): ...

class Query(Serialisable):
    tagname: str
    mdx: Any
    tpls: Any
    __elements__: Any
    def __init__(self, mdx: Any | None = ..., tpls: Any | None = ...) -> None: ...

class QueryCache(Serialisable):
    tagname: str
    count: Any
    query: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., query: Any | None = ...) -> None: ...

class OLAPSet(Serialisable):
    tagname: str
    count: Any
    maxRank: Any
    setDefinition: Any
    sortType: Any
    queryFailed: Any
    tpls: Any
    sortByTuple: Any
    __elements__: Any
    def __init__(
        self,
        count: Any | None = ...,
        maxRank: Any | None = ...,
        setDefinition: Any | None = ...,
        sortType: Any | None = ...,
        queryFailed: Any | None = ...,
        tpls: Any | None = ...,
        sortByTuple: Any | None = ...,
    ) -> None: ...

class OLAPSets(Serialisable):  # type: ignore[misc]
    count: Any
    set: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., set: Any | None = ...) -> None: ...

class PCDSDTCEntries(Serialisable):
    tagname: str
    count: Any
    m: Any
    n: Any
    e: Any
    s: Any
    __elements__: Any
    def __init__(
        self, count: Any | None = ..., m: Any | None = ..., n: Any | None = ..., e: Any | None = ..., s: Any | None = ...
    ) -> None: ...

class TupleCache(Serialisable):
    tagname: str
    entries: Any
    sets: Any
    queryCache: Any
    serverFormats: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        entries: Any | None = ...,
        sets: Any | None = ...,
        queryCache: Any | None = ...,
        serverFormats: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class PCDKPI(Serialisable):
    tagname: str
    uniqueName: Any
    caption: Any
    displayFolder: Any
    measureGroup: Any
    parent: Any
    value: Any
    goal: Any
    status: Any
    trend: Any
    weight: Any
    time: Any
    def __init__(
        self,
        uniqueName: Any | None = ...,
        caption: Any | None = ...,
        displayFolder: Any | None = ...,
        measureGroup: Any | None = ...,
        parent: Any | None = ...,
        value: Any | None = ...,
        goal: Any | None = ...,
        status: Any | None = ...,
        trend: Any | None = ...,
        weight: Any | None = ...,
        time: Any | None = ...,
    ) -> None: ...

class GroupMember(Serialisable):
    tagname: str
    uniqueName: Any
    group: Any
    def __init__(self, uniqueName: Any | None = ..., group: Any | None = ...) -> None: ...

class GroupMembers(Serialisable):  # type: ignore[misc]
    count: Any
    groupMember: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., groupMember: Any | None = ...) -> None: ...

class LevelGroup(Serialisable):
    tagname: str
    name: Any
    uniqueName: Any
    caption: Any
    uniqueParent: Any
    id: Any
    groupMembers: Any
    __elements__: Any
    def __init__(
        self,
        name: Any | None = ...,
        uniqueName: Any | None = ...,
        caption: Any | None = ...,
        uniqueParent: Any | None = ...,
        id: Any | None = ...,
        groupMembers: Any | None = ...,
    ) -> None: ...

class Groups(Serialisable):
    tagname: str
    count: Any
    group: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., group: Any | None = ...) -> None: ...

class GroupLevel(Serialisable):
    tagname: str
    uniqueName: Any
    caption: Any
    user: Any
    customRollUp: Any
    groups: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        uniqueName: Any | None = ...,
        caption: Any | None = ...,
        user: Any | None = ...,
        customRollUp: Any | None = ...,
        groups: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class GroupLevels(Serialisable):  # type: ignore[misc]
    count: Any
    groupLevel: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., groupLevel: Any | None = ...) -> None: ...

class FieldUsage(Serialisable):
    tagname: str
    x: Any
    def __init__(self, x: Any | None = ...) -> None: ...

class FieldsUsage(Serialisable):  # type: ignore[misc]
    count: Any
    fieldUsage: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., fieldUsage: Any | None = ...) -> None: ...

class CacheHierarchy(Serialisable):
    tagname: str
    uniqueName: Any
    caption: Any
    measure: Any
    set: Any
    parentSet: Any
    iconSet: Any
    attribute: Any
    time: Any
    keyAttribute: Any
    defaultMemberUniqueName: Any
    allUniqueName: Any
    allCaption: Any
    dimensionUniqueName: Any
    displayFolder: Any
    measureGroup: Any
    measures: Any
    count: Any
    oneField: Any
    memberValueDatatype: Any
    unbalanced: Any
    unbalancedGroup: Any
    hidden: Any
    fieldsUsage: Any
    groupLevels: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        uniqueName: str = ...,
        caption: Any | None = ...,
        measure: Any | None = ...,
        set: Any | None = ...,
        parentSet: Any | None = ...,
        iconSet: int = ...,
        attribute: Any | None = ...,
        time: Any | None = ...,
        keyAttribute: Any | None = ...,
        defaultMemberUniqueName: Any | None = ...,
        allUniqueName: Any | None = ...,
        allCaption: Any | None = ...,
        dimensionUniqueName: Any | None = ...,
        displayFolder: Any | None = ...,
        measureGroup: Any | None = ...,
        measures: Any | None = ...,
        count: Any | None = ...,
        oneField: Any | None = ...,
        memberValueDatatype: Any | None = ...,
        unbalanced: Any | None = ...,
        unbalancedGroup: Any | None = ...,
        hidden: Any | None = ...,
        fieldsUsage: Any | None = ...,
        groupLevels: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class GroupItems(Serialisable):
    tagname: str
    m: Any
    n: Any
    b: Any
    e: Any
    s: Any
    d: Any
    __elements__: Any
    __attrs__: Any
    def __init__(self, count: Any | None = ..., m=..., n=..., b=..., e=..., s=..., d=...) -> None: ...
    @property
    def count(self): ...

class DiscretePr(Serialisable):
    tagname: str
    count: Any
    x: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., x: Any | None = ...) -> None: ...

class RangePr(Serialisable):
    tagname: str
    autoStart: Any
    autoEnd: Any
    groupBy: Any
    startNum: Any
    endNum: Any
    startDate: Any
    endDate: Any
    groupInterval: Any
    def __init__(
        self,
        autoStart: bool = ...,
        autoEnd: bool = ...,
        groupBy: str = ...,
        startNum: Any | None = ...,
        endNum: Any | None = ...,
        startDate: Any | None = ...,
        endDate: Any | None = ...,
        groupInterval: int = ...,
    ) -> None: ...

class FieldGroup(Serialisable):
    tagname: str
    par: Any
    base: Any
    rangePr: Any
    discretePr: Any
    groupItems: Any
    __elements__: Any
    def __init__(
        self,
        par: Any | None = ...,
        base: Any | None = ...,
        rangePr: Any | None = ...,
        discretePr: Any | None = ...,
        groupItems: Any | None = ...,
    ) -> None: ...

class SharedItems(Serialisable):
    tagname: str
    m: Any
    n: Any
    b: Any
    e: Any
    s: Any
    d: Any
    containsSemiMixedTypes: Any
    containsNonDate: Any
    containsDate: Any
    containsString: Any
    containsBlank: Any
    containsMixedTypes: Any
    containsNumber: Any
    containsInteger: Any
    minValue: Any
    maxValue: Any
    minDate: Any
    maxDate: Any
    longText: Any
    __attrs__: Any
    def __init__(
        self,
        _fields=...,
        containsSemiMixedTypes: Any | None = ...,
        containsNonDate: Any | None = ...,
        containsDate: Any | None = ...,
        containsString: Any | None = ...,
        containsBlank: Any | None = ...,
        containsMixedTypes: Any | None = ...,
        containsNumber: Any | None = ...,
        containsInteger: Any | None = ...,
        minValue: Any | None = ...,
        maxValue: Any | None = ...,
        minDate: Any | None = ...,
        maxDate: Any | None = ...,
        count: Any | None = ...,
        longText: Any | None = ...,
    ) -> None: ...
    @property
    def count(self): ...

class CacheField(Serialisable):
    tagname: str
    sharedItems: Any
    fieldGroup: Any
    mpMap: Any
    extLst: Any
    name: Any
    caption: Any
    propertyName: Any
    serverField: Any
    uniqueList: Any
    numFmtId: Any
    formula: Any
    sqlType: Any
    hierarchy: Any
    level: Any
    databaseField: Any
    mappingCount: Any
    memberPropertyField: Any
    __elements__: Any
    def __init__(
        self,
        sharedItems: Any | None = ...,
        fieldGroup: Any | None = ...,
        mpMap: Any | None = ...,
        extLst: Any | None = ...,
        name: Any | None = ...,
        caption: Any | None = ...,
        propertyName: Any | None = ...,
        serverField: Any | None = ...,
        uniqueList: bool = ...,
        numFmtId: Any | None = ...,
        formula: Any | None = ...,
        sqlType: int = ...,
        hierarchy: int = ...,
        level: int = ...,
        databaseField: bool = ...,
        mappingCount: Any | None = ...,
        memberPropertyField: Any | None = ...,
    ) -> None: ...

class RangeSet(Serialisable):
    tagname: str
    i1: Any
    i2: Any
    i3: Any
    i4: Any
    ref: Any
    name: Any
    sheet: Any
    def __init__(
        self,
        i1: Any | None = ...,
        i2: Any | None = ...,
        i3: Any | None = ...,
        i4: Any | None = ...,
        ref: Any | None = ...,
        name: Any | None = ...,
        sheet: Any | None = ...,
    ) -> None: ...

class PageItem(Serialisable):
    tagname: str
    name: Any
    def __init__(self, name: Any | None = ...) -> None: ...

class Page(Serialisable):
    tagname: str
    pageItem: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., pageItem: Any | None = ...) -> None: ...
    @property
    def count(self): ...

class Consolidation(Serialisable):
    tagname: str
    autoPage: Any
    pages: Any
    rangeSets: Any
    __elements__: Any
    def __init__(self, autoPage: Any | None = ..., pages=..., rangeSets=...) -> None: ...

class WorksheetSource(Serialisable):
    tagname: str
    ref: Any
    name: Any
    sheet: Any
    def __init__(self, ref: Any | None = ..., name: Any | None = ..., sheet: Any | None = ...) -> None: ...

class CacheSource(Serialisable):
    tagname: str
    type: Any
    connectionId: Any
    worksheetSource: Any
    consolidation: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        type: Any | None = ...,
        connectionId: Any | None = ...,
        worksheetSource: Any | None = ...,
        consolidation: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class CacheDefinition(Serialisable):
    mime_type: str
    rel_type: str
    records: Any
    tagname: str
    invalid: Any
    saveData: Any
    refreshOnLoad: Any
    optimizeMemory: Any
    enableRefresh: Any
    refreshedBy: Any
    refreshedDate: Any
    refreshedDateIso: Any
    backgroundQuery: Any
    missingItemsLimit: Any
    createdVersion: Any
    refreshedVersion: Any
    minRefreshableVersion: Any
    recordCount: Any
    upgradeOnRefresh: Any
    tupleCache: Any
    supportSubquery: Any
    supportAdvancedDrill: Any
    cacheSource: Any
    cacheFields: Any
    cacheHierarchies: Any
    kpis: Any
    calculatedItems: Any
    calculatedMembers: Any
    dimensions: Any
    measureGroups: Any
    maps: Any
    extLst: Any
    id: Any
    __elements__: Any
    def __init__(
        self,
        invalid: Any | None = ...,
        saveData: Any | None = ...,
        refreshOnLoad: Any | None = ...,
        optimizeMemory: Any | None = ...,
        enableRefresh: Any | None = ...,
        refreshedBy: Any | None = ...,
        refreshedDate: Any | None = ...,
        refreshedDateIso: Any | None = ...,
        backgroundQuery: Any | None = ...,
        missingItemsLimit: Any | None = ...,
        createdVersion: Any | None = ...,
        refreshedVersion: Any | None = ...,
        minRefreshableVersion: Any | None = ...,
        recordCount: Any | None = ...,
        upgradeOnRefresh: Any | None = ...,
        tupleCache: Any | None = ...,
        supportSubquery: Any | None = ...,
        supportAdvancedDrill: Any | None = ...,
        cacheSource: Any | None = ...,
        cacheFields=...,
        cacheHierarchies=...,
        kpis=...,
        calculatedItems=...,
        calculatedMembers=...,
        dimensions=...,
        measureGroups=...,
        maps=...,
        extLst: Any | None = ...,
        id: Any | None = ...,
    ) -> None: ...
    def to_tree(self): ...
    @property
    def path(self): ...



from .prettytable import (
    ALL as ALL,
    DEFAULT as DEFAULT,
    FRAME as FRAME,
    HEADER as HEADER,
    MARKDOWN as MARKDOWN,
    MSWORD_FRIENDLY as MSWORD_FRIENDLY,
    NONE as NONE,
    ORGMODE as ORGMODE,
    PLAIN_COLUMNS as PLAIN_COLUMNS,
    RANDOM as RANDOM,
    PrettyTable as PrettyTable,
    TableHandler as TableHandler,
    from_csv as from_csv,
    from_db_cursor as from_db_cursor,
    from_html as from_html,
    from_html_one as from_html_one,
    from_json as from_json,
)



from typing import Any

from braintree.attribute_getter import AttributeGetter as AttributeGetter
from braintree.merchant_account.address_details import AddressDetails as AddressDetails

class IndividualDetails(AttributeGetter):
    detail_list: Any
    address_details: Any
    def __init__(self, attributes) -> None: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

horizontal_alignments: Any
vertical_aligments: Any

class Alignment(Serialisable):
    tagname: str
    __fields__: Any
    horizontal: Any
    vertical: Any
    textRotation: Any
    text_rotation: Any
    wrapText: Any
    wrap_text: Any
    shrinkToFit: Any
    shrink_to_fit: Any
    indent: Any
    relativeIndent: Any
    justifyLastLine: Any
    readingOrder: Any
    def __init__(
        self,
        horizontal: Any | None = ...,
        vertical: Any | None = ...,
        textRotation: int = ...,
        wrapText: Any | None = ...,
        shrinkToFit: Any | None = ...,
        indent: int = ...,
        relativeIndent: int = ...,
        justifyLastLine: Any | None = ...,
        readingOrder: int = ...,
        text_rotation: Any | None = ...,
        wrap_text: Any | None = ...,
        shrink_to_fit: Any | None = ...,
        mergeCell: Any | None = ...,
    ) -> None: ...
    def __iter__(self): ...



import abc

from ..util import ABC

class ConnectionCharacteristic(ABC, metaclass=abc.ABCMeta):
    transactional: bool
    @abc.abstractmethod
    def reset_characteristic(self, dialect, dbapi_conn): ...
    @abc.abstractmethod
    def set_characteristic(self, dialect, dbapi_conn, value): ...
    @abc.abstractmethod
    def get_characteristic(self, dialect, dbapi_conn): ...

class IsolationLevelCharacteristic(ConnectionCharacteristic):
    transactional: bool
    def reset_characteristic(self, dialect, dbapi_conn) -> None: ...
    def set_characteristic(self, dialect, dbapi_conn, value) -> None: ...
    def get_characteristic(self, dialect, dbapi_conn): ...



from distutils import log as log
from distutils.ccompiler import CCompiler
from distutils.core import Command as Command
from distutils.errors import DistutilsExecError as DistutilsExecError
from distutils.sysconfig import customize_compiler as customize_compiler
from typing import Pattern, Sequence

LANG_EXT: dict[str, str]

class config(Command):
    description: str = ...
    # Tuple is full name, short name, description
    user_options: Sequence[tuple[str, str | None, str]] = ...
    compiler: str | CCompiler | None = ...
    cc: str | None = ...
    include_dirs: Sequence[str] | None = ...
    libraries: Sequence[str] | None = ...
    library_dirs: Sequence[str] | None = ...
    noisy: int = ...
    dump_source: int = ...
    temp_files: Sequence[str] = ...
    def initialize_options(self) -> None: ...
    def finalize_options(self) -> None: ...
    def run(self) -> None: ...
    def try_cpp(
        self,
        body: str | None = ...,
        headers: Sequence[str] | None = ...,
        include_dirs: Sequence[str] | None = ...,
        lang: str = ...,
    ) -> bool: ...
    def search_cpp(
        self,
        pattern: Pattern[str] | str,
        body: str | None = ...,
        headers: Sequence[str] | None = ...,
        include_dirs: Sequence[str] | None = ...,
        lang: str = ...,
    ) -> bool: ...
    def try_compile(
        self, body: str, headers: Sequence[str] | None = ..., include_dirs: Sequence[str] | None = ..., lang: str = ...
    ) -> bool: ...
    def try_link(
        self,
        body: str,
        headers: Sequence[str] | None = ...,
        include_dirs: Sequence[str] | None = ...,
        libraries: Sequence[str] | None = ...,
        library_dirs: Sequence[str] | None = ...,
        lang: str = ...,
    ) -> bool: ...
    def try_run(
        self,
        body: str,
        headers: Sequence[str] | None = ...,
        include_dirs: Sequence[str] | None = ...,
        libraries: Sequence[str] | None = ...,
        library_dirs: Sequence[str] | None = ...,
        lang: str = ...,
    ) -> bool: ...
    def check_func(
        self,
        func: str,
        headers: Sequence[str] | None = ...,
        include_dirs: Sequence[str] | None = ...,
        libraries: Sequence[str] | None = ...,
        library_dirs: Sequence[str] | None = ...,
        decl: int = ...,
        call: int = ...,
    ) -> bool: ...
    def check_lib(
        self,
        library: str,
        library_dirs: Sequence[str] | None = ...,
        headers: Sequence[str] | None = ...,
        include_dirs: Sequence[str] | None = ...,
        other_libraries: list[str] = ...,
    ) -> bool: ...
    def check_header(
        self, header: str, include_dirs: Sequence[str] | None = ..., library_dirs: Sequence[str] | None = ..., lang: str = ...
    ) -> bool: ...

def dump_file(filename: str, head: str | None = ...) -> None: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class HierarchyUsage(Serialisable):
    tagname: str
    hierarchyUsage: Any
    def __init__(self, hierarchyUsage: Any | None = ...) -> None: ...

class ColHierarchiesUsage(Serialisable):
    tagname: str
    colHierarchyUsage: Any
    __elements__: Any
    __attrs__: Any
    def __init__(self, count: Any | None = ..., colHierarchyUsage=...) -> None: ...
    @property
    def count(self): ...

class RowHierarchiesUsage(Serialisable):
    tagname: str
    rowHierarchyUsage: Any
    __elements__: Any
    __attrs__: Any
    def __init__(self, count: Any | None = ..., rowHierarchyUsage=...) -> None: ...
    @property
    def count(self): ...

class PivotFilter(Serialisable):
    tagname: str
    fld: Any
    mpFld: Any
    type: Any
    evalOrder: Any
    id: Any
    iMeasureHier: Any
    iMeasureFld: Any
    name: Any
    description: Any
    stringValue1: Any
    stringValue2: Any
    autoFilter: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        fld: Any | None = ...,
        mpFld: Any | None = ...,
        type: Any | None = ...,
        evalOrder: Any | None = ...,
        id: Any | None = ...,
        iMeasureHier: Any | None = ...,
        iMeasureFld: Any | None = ...,
        name: Any | None = ...,
        description: Any | None = ...,
        stringValue1: Any | None = ...,
        stringValue2: Any | None = ...,
        autoFilter: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class PivotFilters(Serialisable):  # type: ignore[misc]
    count: Any
    filter: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., filter: Any | None = ...) -> None: ...

class PivotTableStyle(Serialisable):
    tagname: str
    name: Any
    showRowHeaders: Any
    showColHeaders: Any
    showRowStripes: Any
    showColStripes: Any
    showLastColumn: Any
    def __init__(
        self,
        name: Any | None = ...,
        showRowHeaders: Any | None = ...,
        showColHeaders: Any | None = ...,
        showRowStripes: Any | None = ...,
        showColStripes: Any | None = ...,
        showLastColumn: Any | None = ...,
    ) -> None: ...

class MemberList(Serialisable):
    tagname: str
    level: Any
    member: Any
    __elements__: Any
    def __init__(self, count: Any | None = ..., level: Any | None = ..., member=...) -> None: ...
    @property
    def count(self): ...

class MemberProperty(Serialisable):
    tagname: str
    name: Any
    showCell: Any
    showTip: Any
    showAsCaption: Any
    nameLen: Any
    pPos: Any
    pLen: Any
    level: Any
    field: Any
    def __init__(
        self,
        name: Any | None = ...,
        showCell: Any | None = ...,
        showTip: Any | None = ...,
        showAsCaption: Any | None = ...,
        nameLen: Any | None = ...,
        pPos: Any | None = ...,
        pLen: Any | None = ...,
        level: Any | None = ...,
        field: Any | None = ...,
    ) -> None: ...

class PivotHierarchy(Serialisable):
    tagname: str
    outline: Any
    multipleItemSelectionAllowed: Any
    subtotalTop: Any
    showInFieldList: Any
    dragToRow: Any
    dragToCol: Any
    dragToPage: Any
    dragToData: Any
    dragOff: Any
    includeNewItemsInFilter: Any
    caption: Any
    mps: Any
    members: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        outline: Any | None = ...,
        multipleItemSelectionAllowed: Any | None = ...,
        subtotalTop: Any | None = ...,
        showInFieldList: Any | None = ...,
        dragToRow: Any | None = ...,
        dragToCol: Any | None = ...,
        dragToPage: Any | None = ...,
        dragToData: Any | None = ...,
        dragOff: Any | None = ...,
        includeNewItemsInFilter: Any | None = ...,
        caption: Any | None = ...,
        mps=...,
        members: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class Reference(Serialisable):
    tagname: str
    field: Any
    selected: Any
    byPosition: Any
    relative: Any
    defaultSubtotal: Any
    sumSubtotal: Any
    countASubtotal: Any
    avgSubtotal: Any
    maxSubtotal: Any
    minSubtotal: Any
    productSubtotal: Any
    countSubtotal: Any
    stdDevSubtotal: Any
    stdDevPSubtotal: Any
    varSubtotal: Any
    varPSubtotal: Any
    x: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        field: Any | None = ...,
        count: Any | None = ...,
        selected: Any | None = ...,
        byPosition: Any | None = ...,
        relative: Any | None = ...,
        defaultSubtotal: Any | None = ...,
        sumSubtotal: Any | None = ...,
        countASubtotal: Any | None = ...,
        avgSubtotal: Any | None = ...,
        maxSubtotal: Any | None = ...,
        minSubtotal: Any | None = ...,
        productSubtotal: Any | None = ...,
        countSubtotal: Any | None = ...,
        stdDevSubtotal: Any | None = ...,
        stdDevPSubtotal: Any | None = ...,
        varSubtotal: Any | None = ...,
        varPSubtotal: Any | None = ...,
        x: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...
    @property
    def count(self): ...

class PivotArea(Serialisable):
    tagname: str
    references: Any
    extLst: Any
    field: Any
    type: Any
    dataOnly: Any
    labelOnly: Any
    grandRow: Any
    grandCol: Any
    cacheIndex: Any
    outline: Any
    offset: Any
    collapsedLevelsAreSubtotals: Any
    axis: Any
    fieldPosition: Any
    __elements__: Any
    def __init__(
        self,
        references=...,
        extLst: Any | None = ...,
        field: Any | None = ...,
        type: str = ...,
        dataOnly: bool = ...,
        labelOnly: Any | None = ...,
        grandRow: Any | None = ...,
        grandCol: Any | None = ...,
        cacheIndex: Any | None = ...,
        outline: bool = ...,
        offset: Any | None = ...,
        collapsedLevelsAreSubtotals: Any | None = ...,
        axis: Any | None = ...,
        fieldPosition: Any | None = ...,
    ) -> None: ...

class ChartFormat(Serialisable):
    tagname: str
    chart: Any
    format: Any
    series: Any
    pivotArea: Any
    __elements__: Any
    def __init__(
        self, chart: Any | None = ..., format: Any | None = ..., series: Any | None = ..., pivotArea: Any | None = ...
    ) -> None: ...

class ConditionalFormat(Serialisable):
    tagname: str
    scope: Any
    type: Any
    priority: Any
    pivotAreas: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        scope: Any | None = ...,
        type: Any | None = ...,
        priority: Any | None = ...,
        pivotAreas=...,
        extLst: Any | None = ...,
    ) -> None: ...

class Format(Serialisable):
    tagname: str
    action: Any
    dxfId: Any
    pivotArea: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self, action: str = ..., dxfId: Any | None = ..., pivotArea: Any | None = ..., extLst: Any | None = ...
    ) -> None: ...

class DataField(Serialisable):
    tagname: str
    name: Any
    fld: Any
    subtotal: Any
    showDataAs: Any
    baseField: Any
    baseItem: Any
    numFmtId: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        name: Any | None = ...,
        fld: Any | None = ...,
        subtotal: str = ...,
        showDataAs: str = ...,
        baseField: int = ...,
        baseItem: int = ...,
        numFmtId: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class PageField(Serialisable):
    tagname: str
    fld: Any
    item: Any
    hier: Any
    name: Any
    cap: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        fld: Any | None = ...,
        item: Any | None = ...,
        hier: Any | None = ...,
        name: Any | None = ...,
        cap: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class RowColItem(Serialisable):
    tagname: str
    t: Any
    r: Any
    i: Any
    x: Any
    __elements__: Any
    def __init__(self, t: str = ..., r: int = ..., i: int = ..., x=...) -> None: ...

class RowColField(Serialisable):
    tagname: str
    x: Any
    def __init__(self, x: Any | None = ...) -> None: ...

class AutoSortScope(Serialisable):  # type: ignore[misc]
    pivotArea: Any
    __elements__: Any
    def __init__(self, pivotArea: Any | None = ...) -> None: ...

class FieldItem(Serialisable):
    tagname: str
    n: Any
    t: Any
    h: Any
    s: Any
    sd: Any
    f: Any
    m: Any
    c: Any
    x: Any
    d: Any
    e: Any
    def __init__(
        self,
        n: Any | None = ...,
        t: str = ...,
        h: Any | None = ...,
        s: Any | None = ...,
        sd: bool = ...,
        f: Any | None = ...,
        m: Any | None = ...,
        c: Any | None = ...,
        x: Any | None = ...,
        d: Any | None = ...,
        e: Any | None = ...,
    ) -> None: ...

class PivotField(Serialisable):
    tagname: str
    items: Any
    autoSortScope: Any
    extLst: Any
    name: Any
    axis: Any
    dataField: Any
    subtotalCaption: Any
    showDropDowns: Any
    hiddenLevel: Any
    uniqueMemberProperty: Any
    compact: Any
    allDrilled: Any
    numFmtId: Any
    outline: Any
    subtotalTop: Any
    dragToRow: Any
    dragToCol: Any
    multipleItemSelectionAllowed: Any
    dragToPage: Any
    dragToData: Any
    dragOff: Any
    showAll: Any
    insertBlankRow: Any
    serverField: Any
    insertPageBreak: Any
    autoShow: Any
    topAutoShow: Any
    hideNewItems: Any
    measureFilter: Any
    includeNewItemsInFilter: Any
    itemPageCount: Any
    sortType: Any
    dataSourceSort: Any
    nonAutoSortDefault: Any
    rankBy: Any
    defaultSubtotal: Any
    sumSubtotal: Any
    countASubtotal: Any
    avgSubtotal: Any
    maxSubtotal: Any
    minSubtotal: Any
    productSubtotal: Any
    countSubtotal: Any
    stdDevSubtotal: Any
    stdDevPSubtotal: Any
    varSubtotal: Any
    varPSubtotal: Any
    showPropCell: Any
    showPropTip: Any
    showPropAsCaption: Any
    defaultAttributeDrillState: Any
    __elements__: Any
    def __init__(
        self,
        items=...,
        autoSortScope: Any | None = ...,
        name: Any | None = ...,
        axis: Any | None = ...,
        dataField: Any | None = ...,
        subtotalCaption: Any | None = ...,
        showDropDowns: bool = ...,
        hiddenLevel: Any | None = ...,
        uniqueMemberProperty: Any | None = ...,
        compact: bool = ...,
        allDrilled: Any | None = ...,
        numFmtId: Any | None = ...,
        outline: bool = ...,
        subtotalTop: bool = ...,
        dragToRow: bool = ...,
        dragToCol: bool = ...,
        multipleItemSelectionAllowed: Any | None = ...,
        dragToPage: bool = ...,
        dragToData: bool = ...,
        dragOff: bool = ...,
        showAll: bool = ...,
        insertBlankRow: Any | None = ...,
        serverField: Any | None = ...,
        insertPageBreak: Any | None = ...,
        autoShow: Any | None = ...,
        topAutoShow: bool = ...,
        hideNewItems: Any | None = ...,
        measureFilter: Any | None = ...,
        includeNewItemsInFilter: Any | None = ...,
        itemPageCount: int = ...,
        sortType: str = ...,
        dataSourceSort: Any | None = ...,
        nonAutoSortDefault: Any | None = ...,
        rankBy: Any | None = ...,
        defaultSubtotal: bool = ...,
        sumSubtotal: Any | None = ...,
        countASubtotal: Any | None = ...,
        avgSubtotal: Any | None = ...,
        maxSubtotal: Any | None = ...,
        minSubtotal: Any | None = ...,
        productSubtotal: Any | None = ...,
        countSubtotal: Any | None = ...,
        stdDevSubtotal: Any | None = ...,
        stdDevPSubtotal: Any | None = ...,
        varSubtotal: Any | None = ...,
        varPSubtotal: Any | None = ...,
        showPropCell: Any | None = ...,
        showPropTip: Any | None = ...,
        showPropAsCaption: Any | None = ...,
        defaultAttributeDrillState: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class Location(Serialisable):
    tagname: str
    ref: Any
    firstHeaderRow: Any
    firstDataRow: Any
    firstDataCol: Any
    rowPageCount: Any
    colPageCount: Any
    def __init__(
        self,
        ref: Any | None = ...,
        firstHeaderRow: Any | None = ...,
        firstDataRow: Any | None = ...,
        firstDataCol: Any | None = ...,
        rowPageCount: Any | None = ...,
        colPageCount: Any | None = ...,
    ) -> None: ...

class TableDefinition(Serialisable):
    mime_type: str
    rel_type: str
    tagname: str
    cache: Any
    name: Any
    cacheId: Any
    dataOnRows: Any
    dataPosition: Any
    dataCaption: Any
    grandTotalCaption: Any
    errorCaption: Any
    showError: Any
    missingCaption: Any
    showMissing: Any
    pageStyle: Any
    pivotTableStyle: Any
    vacatedStyle: Any
    tag: Any
    updatedVersion: Any
    minRefreshableVersion: Any
    asteriskTotals: Any
    showItems: Any
    editData: Any
    disableFieldList: Any
    showCalcMbrs: Any
    visualTotals: Any
    showMultipleLabel: Any
    showDataDropDown: Any
    showDrill: Any
    printDrill: Any
    showMemberPropertyTips: Any
    showDataTips: Any
    enableWizard: Any
    enableDrill: Any
    enableFieldProperties: Any
    preserveFormatting: Any
    useAutoFormatting: Any
    pageWrap: Any
    pageOverThenDown: Any
    subtotalHiddenItems: Any
    rowGrandTotals: Any
    colGrandTotals: Any
    fieldPrintTitles: Any
    itemPrintTitles: Any
    mergeItem: Any
    showDropZones: Any
    createdVersion: Any
    indent: Any
    showEmptyRow: Any
    showEmptyCol: Any
    showHeaders: Any
    compact: Any
    outline: Any
    outlineData: Any
    compactData: Any
    published: Any
    gridDropZones: Any
    immersive: Any
    multipleFieldFilters: Any
    chartFormat: Any
    rowHeaderCaption: Any
    colHeaderCaption: Any
    fieldListSortAscending: Any
    mdxSubqueries: Any
    customListSort: Any
    autoFormatId: Any
    applyNumberFormats: Any
    applyBorderFormats: Any
    applyFontFormats: Any
    applyPatternFormats: Any
    applyAlignmentFormats: Any
    applyWidthHeightFormats: Any
    location: Any
    pivotFields: Any
    rowFields: Any
    rowItems: Any
    colFields: Any
    colItems: Any
    pageFields: Any
    dataFields: Any
    formats: Any
    conditionalFormats: Any
    chartFormats: Any
    pivotHierarchies: Any
    pivotTableStyleInfo: Any
    filters: Any
    rowHierarchiesUsage: Any
    colHierarchiesUsage: Any
    extLst: Any
    id: Any
    __elements__: Any
    def __init__(
        self,
        name: Any | None = ...,
        cacheId: Any | None = ...,
        dataOnRows: bool = ...,
        dataPosition: Any | None = ...,
        dataCaption: Any | None = ...,
        grandTotalCaption: Any | None = ...,
        errorCaption: Any | None = ...,
        showError: bool = ...,
        missingCaption: Any | None = ...,
        showMissing: bool = ...,
        pageStyle: Any | None = ...,
        pivotTableStyle: Any | None = ...,
        vacatedStyle: Any | None = ...,
        tag: Any | None = ...,
        updatedVersion: int = ...,
        minRefreshableVersion: int = ...,
        asteriskTotals: bool = ...,
        showItems: bool = ...,
        editData: bool = ...,
        disableFieldList: bool = ...,
        showCalcMbrs: bool = ...,
        visualTotals: bool = ...,
        showMultipleLabel: bool = ...,
        showDataDropDown: bool = ...,
        showDrill: bool = ...,
        printDrill: bool = ...,
        showMemberPropertyTips: bool = ...,
        showDataTips: bool = ...,
        enableWizard: bool = ...,
        enableDrill: bool = ...,
        enableFieldProperties: bool = ...,
        preserveFormatting: bool = ...,
        useAutoFormatting: bool = ...,
        pageWrap: int = ...,
        pageOverThenDown: bool = ...,
        subtotalHiddenItems: bool = ...,
        rowGrandTotals: bool = ...,
        colGrandTotals: bool = ...,
        fieldPrintTitles: bool = ...,
        itemPrintTitles: bool = ...,
        mergeItem: bool = ...,
        showDropZones: bool = ...,
        createdVersion: int = ...,
        indent: int = ...,
        showEmptyRow: bool = ...,
        showEmptyCol: bool = ...,
        showHeaders: bool = ...,
        compact: bool = ...,
        outline: bool = ...,
        outlineData: bool = ...,
        compactData: bool = ...,
        published: bool = ...,
        gridDropZones: bool = ...,
        immersive: bool = ...,
        multipleFieldFilters: Any | None = ...,
        chartFormat: int = ...,
        rowHeaderCaption: Any | None = ...,
        colHeaderCaption: Any | None = ...,
        fieldListSortAscending: Any | None = ...,
        mdxSubqueries: Any | None = ...,
        customListSort: Any | None = ...,
        autoFormatId: Any | None = ...,
        applyNumberFormats: bool = ...,
        applyBorderFormats: bool = ...,
        applyFontFormats: bool = ...,
        applyPatternFormats: bool = ...,
        applyAlignmentFormats: bool = ...,
        applyWidthHeightFormats: bool = ...,
        location: Any | None = ...,
        pivotFields=...,
        rowFields=...,
        rowItems=...,
        colFields=...,
        colItems=...,
        pageFields=...,
        dataFields=...,
        formats=...,
        conditionalFormats=...,
        chartFormats=...,
        pivotHierarchies=...,
        pivotTableStyleInfo: Any | None = ...,
        filters=...,
        rowHierarchiesUsage: Any | None = ...,
        colHierarchiesUsage: Any | None = ...,
        extLst: Any | None = ...,
        id: Any | None = ...,
    ) -> None: ...
    def to_tree(self): ...
    @property
    def path(self): ...



from typing import Any, Generic, Iterable, Iterator, Mapping, MutableMapping, TypeVar

_VT = TypeVar("_VT")

class CaseInsensitiveDict(MutableMapping[str, _VT], Generic[_VT]):
    def __init__(self, data: Mapping[str, _VT] | Iterable[tuple[str, _VT]] | None = ..., **kwargs: _VT) -> None: ...
    def lower_items(self) -> Iterator[tuple[str, _VT]]: ...
    def __setitem__(self, key: str, value: _VT) -> None: ...
    def __getitem__(self, key: str) -> _VT: ...
    def __delitem__(self, key: str) -> None: ...
    def __iter__(self) -> Iterator[str]: ...
    def __len__(self) -> int: ...
    def copy(self) -> CaseInsensitiveDict[_VT]: ...

class LookupDict(dict[str, _VT]):
    name: Any
    def __init__(self, name: Any = ...) -> None: ...
    def __getitem__(self, key: str) -> _VT | None: ...  # type: ignore[override]
    def __getattr__(self, attr: str) -> _VT: ...
    def __setattr__(self, __attr: str, __value: _VT) -> None: ...



from typing import Any

Cc: str
Cf: str
Cn: str
Co: str
Cs: str
Ll: str
Lm: str
Lo: str
Lt: str
Lu: str
Mc: str
Me: str
Mn: str
Nd: str
Nl: str
No: str
Pc: str
Pd: str
Pe: str
Pf: str
Pi: str
Po: str
Ps: str
Sc: str
Sk: str
Sm: str
So: str
Zl: str
Zp: str
Zs: str
xid_continue: str
xid_start: str
cats: Any

def combine(*args): ...
def allexcept(*args): ...



import sys
from _typeshed import Self
from typing import Any, Generic, Iterable, Iterator, MutableSet, TypeVar

if sys.version_info >= (3, 9):
    from types import GenericAlias

__all__ = ["WeakSet"]

_S = TypeVar("_S")
_T = TypeVar("_T")

class WeakSet(MutableSet[_T], Generic[_T]):
    def __init__(self, data: Iterable[_T] | None = ...) -> None: ...
    def add(self, item: _T) -> None: ...
    def clear(self) -> None: ...
    def discard(self, item: _T) -> None: ...
    def copy(self: Self) -> Self: ...
    def pop(self) -> _T: ...
    def remove(self, item: _T) -> None: ...
    def update(self, other: Iterable[_T]) -> None: ...
    def __contains__(self, item: object) -> bool: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[_T]: ...
    def __ior__(self: Self, other: Iterable[_T]) -> Self: ...  # type: ignore[override,misc]
    def difference(self: Self, other: Iterable[_T]) -> Self: ...
    def __sub__(self: Self, other: Iterable[Any]) -> Self: ...
    def difference_update(self, other: Iterable[Any]) -> None: ...
    def __isub__(self: Self, other: Iterable[Any]) -> Self: ...
    def intersection(self: Self, other: Iterable[_T]) -> Self: ...
    def __and__(self: Self, other: Iterable[Any]) -> Self: ...
    def intersection_update(self, other: Iterable[Any]) -> None: ...
    def __iand__(self: Self, other: Iterable[Any]) -> Self: ...
    def issubset(self, other: Iterable[_T]) -> bool: ...
    def __le__(self, other: Iterable[_T]) -> bool: ...
    def __lt__(self, other: Iterable[_T]) -> bool: ...
    def issuperset(self, other: Iterable[_T]) -> bool: ...
    def __ge__(self, other: Iterable[_T]) -> bool: ...
    def __gt__(self, other: Iterable[_T]) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def symmetric_difference(self, other: Iterable[_S]) -> WeakSet[_S | _T]: ...
    def __xor__(self, other: Iterable[_S]) -> WeakSet[_S | _T]: ...
    def symmetric_difference_update(self, other: Iterable[_T]) -> None: ...
    def __ixor__(self: Self, other: Iterable[_T]) -> Self: ...  # type: ignore[override,misc]
    def union(self, other: Iterable[_S]) -> WeakSet[_S | _T]: ...
    def __or__(self, other: Iterable[_S]) -> WeakSet[_S | _T]: ...
    def isdisjoint(self, other: Iterable[_T]) -> bool: ...
    if sys.version_info >= (3, 9):
        def __class_getitem__(cls, item: Any) -> GenericAlias: ...



"""Public interface to top-level pytype functions."""

import contextlib
import logging
import os
import sys
import traceback

import libcst

from pytype import __version__
from pytype import analyze
from pytype import config
from pytype import constant_folding
from pytype import errors
from pytype import load_pytd
from pytype import utils
from pytype.directors import directors
from pytype.pyc import pyc
from pytype.pyi import parser
from pytype.pytd import builtin_stubs as pytd_builtins
from pytype.pytd import optimize
from pytype.pytd import pytd_utils
from pytype.pytd import serialize_ast
from pytype.pytd import visitors


log = logging.getLogger(__name__)


# Webpage explaining the pytype error codes
ERROR_DOC_URL = "https://google.github.io/pytype/errors.html"


def read_source_file(input_filename, open_function=open):
  try:
    with open_function(input_filename, "r", encoding="utf8") as fi:
      return fi.read()
  except OSError as e:
    raise utils.UsageError(f"Could not load input file {input_filename}") from e


def _set_verbosity_from(posarg):
  """Decorator to set the verbosity for a function that takes an options arg.

  Assumes that the function has an argument named `options` that is a
  config.Options object.

  Arguments:
    posarg: The index of `options` in the positional arguments.

  Returns:
    The decorator.
  """
  def decorator(f):
    def wrapper(*args, **kwargs):
      options = kwargs.get("options", args[posarg])
      with config.verbosity_from(options):
        return f(*args, **kwargs)
    return wrapper
  return decorator


@_set_verbosity_from(posarg=2)
def _call(analyze_types, src, options, loader):
  """Helper function to call analyze.check/infer_types."""
  # 'deep' tells the analyzer whether to analyze functions not called from main.
  deep = not options.main_only
  loader = loader or load_pytd.create_loader(options)
  return analyze_types(
      src=src,
      filename=options.input,
      options=options,
      loader=loader,
      deep=deep)


def check_py(src, options=None, loader=None):
  """Check the types of a string of source code."""
  options = options or config.Options.create()
  with config.verbosity_from(options):
    ret = _call(analyze.check_types, src, options, loader)
  return ret.errorlog


def generate_pyi(src, options=None, loader=None):
  """Run the inferencer on a string of source code, producing output.

  Args:
    src: The source code.
    options: config.Options object.
    loader: A load_pytd.Loader instance.

  Returns:
    A tuple, (errors.ErrorLog, PYI Ast as string, TypeDeclUnit).

  Raises:
    CompileError: If we couldn't parse the input file.
    UsageError: If the input filepath is invalid.
  """
  options = options or config.Options.create()
  with config.verbosity_from(options):
    ret = _call(analyze.infer_types, src, options, loader)
    mod = ret.ast
    mod.Visit(visitors.VerifyVisitor())
    mod = optimize.Optimize(mod,
                            ret.builtins,
                            lossy=False,
                            use_abcs=False,
                            max_union=7,
                            remove_mutable=False)
    mod = pytd_utils.CanonicalOrdering(mod)
    result = pytd_utils.Print(mod)
    log.info("=========== pyi optimized =============")
    log.info("\n%s", result)
    log.info("========================================")

  result += "\n"
  if options.quick:
    result = "# (generated with --quick)\n\n" + result
  return ret.errorlog, result, mod


@_set_verbosity_from(posarg=0)
def check_or_generate_pyi(options, loader=None):
  """Returns generated errors and result pyi or None if it's only check.

  Args:
    options: config.Options object.
    loader: load_pytd.Loader object.

  Returns:
    A tuple, (errors.ErrorLog, PYI Ast as string or None, AST or None).
  """

  errorlog = errors.ErrorLog()
  result = pytd_builtins.DEFAULT_SRC
  ast = pytd_builtins.GetDefaultAst(
      parser.PyiOptions.from_toplevel_options(options))
  try:
    src = read_source_file(options.input, options.open_function)
    if options.check:
      return check_py(src=src, options=options, loader=loader), None, None
    else:
      errorlog, result, ast = generate_pyi(
          src=src, options=options, loader=loader)
  except utils.UsageError:
    raise
  except pyc.CompileError as e:
    errorlog.python_compiler_error(options.input, e.lineno, e.error)
  except constant_folding.ConstantError as e:
    errorlog.python_compiler_error(options.input, e.lineno, e.message)
  except IndentationError as e:
    errorlog.python_compiler_error(options.input, e.lineno, e.msg)
  except libcst.ParserSyntaxError as e:
    # TODO(rechen): We can get rid of this branch once we delete
    # directors.parser_libcst.
    errorlog.python_compiler_error(options.input, e.raw_line, e.message)
  except SyntaxError as e:
    errorlog.python_compiler_error(options.input, e.lineno, e.msg)
  except directors.SkipFileError:
    result += "# skip-file found, file not analyzed"
  except Exception as e:  # pylint: disable=broad-except
    if options.nofail:
      log.warning("***Caught exception: %s", str(e), exc_info=True)
      if not options.check:
        result += (
            "# Caught error in pytype: " + str(e).replace("\n", "\n#")
            + "\n# " + "\n# ".join(traceback.format_exc().splitlines()))
    else:
      e.args = (
          str(utils.message(e)) + f"\nFile: {options.input}",) + e.args[1:]
      raise

  return (errorlog, None, None) if options.check else (errorlog, result, ast)


def _write_pyi_output(options, contents, filename):
  assert filename
  if filename == "-":
    sys.stdout.write(contents)
  else:
    log.info("write pyi %r => %r", options.input, filename)
    with options.open_function(filename, "w") as fi:
      fi.write(contents)


@_set_verbosity_from(posarg=0)
def process_one_file(options):
  """Check a .py file or generate a .pyi for it, according to options.

  Args:
    options: config.Options object.

  Returns:
    An error code (0 means no error).
  """

  log.info("Process %s => %s", options.input, options.output)
  loader = load_pytd.create_loader(options)
  try:
    errorlog, result, ast = check_or_generate_pyi(options, loader)
  except utils.UsageError as e:
    logging.error("Usage error: %s\n", utils.message(e))
    return 1

  if not options.check:
    if options.pickle_output:
      pyi_output = options.verify_pickle
    else:
      pyi_output = options.output
    # Write out the pyi file.
    if pyi_output:
      _write_pyi_output(options, result, pyi_output)
    # Write out the pickle file.
    if options.pickle_output:
      log.info("write pickle %r => %r", options.input, options.output)
      write_pickle(ast, options, loader)
  exit_status = handle_errors(errorlog, options)

  # If we have set return_success, set exit_status to 0 after the regular error
  # handler has been called.
  if options.return_success:
    exit_status = 0

  # Touch output file upon success.
  if options.touch and not exit_status:
    with options.open_function(options.touch, "a"):
      os.utime(options.touch, None)
  return exit_status


@_set_verbosity_from(posarg=1)
def write_pickle(ast, options, loader=None):
  """Dump a pickle of the ast to a file."""
  loader = loader or load_pytd.create_loader(options)
  try:
    ast = serialize_ast.PrepareForExport(options.module_name, ast, loader)
  except parser.ParseError as e:
    if options.nofail:
      ast = serialize_ast.PrepareForExport(
          options.module_name, loader.get_default_ast(), loader)
      log.warning("***Caught exception: %s", str(e), exc_info=True)
    else:
      raise
  if options.verify_pickle:
    ast1 = ast.Visit(visitors.LateTypeToClassType())
    ast1 = ast1.Visit(visitors.ClearClassPointers())
    ast2 = loader.load_file(options.module_name, options.verify_pickle)
    ast2 = ast2.Visit(visitors.ClearClassPointers())
    if not pytd_utils.ASTeq(ast1, ast2):
      raise AssertionError()
  serialize_ast.StoreAst(ast, options.output, options.open_function,
                         is_package=options.module_name.endswith(".__init__"),
                         metadata=options.pickle_metadata)


def print_error_doc_url(errorlog):
  names = {e.name for e in errorlog}
  if names:
    doclink = f"\nFor more details, see {ERROR_DOC_URL}"
    if len(names) == 1:
      doclink += "#" + names.pop()
    print(doclink, file=sys.stderr)


@_set_verbosity_from(posarg=1)
def handle_errors(errorlog, options):
  """Handle the errorlog according to the given options."""
  if not options.report_errors:
    return 0

  if options.output_errors_csv:
    errorlog.print_to_csv_file(options.output_errors_csv, options.open_function)
    return 0  # Command is successful regardless of errors.

  errorlog.print_to_stderr(color=options.color)
  print_error_doc_url(errorlog)

  return 1 if errorlog.has_error() else 0  # exit code


@_set_verbosity_from(posarg=0)
def parse_pyi(options):
  """Tries parsing a PYI file."""
  loader = load_pytd.create_loader(options)
  ast = loader.load_file(options.module_name, options.input)
  ast = loader.finish_and_verify_ast(ast)
  if options.output:
    result = "# Internal AST parsed and postprocessed from {}\n\n{}".format(
        options.input, pytd_utils.Print(ast))
    _write_pyi_output(options, result, options.output)
  return ast


def get_pytype_version():
  return __version__.__version__


@contextlib.contextmanager
def wrap_pytype_exceptions(exception_type, filename=""):
  """Catch pytype errors and reraise them as a single exception type.

  NOTE: This will also wrap non-pytype errors thrown within the body of the
  code block; it is therefore recommended to use this to wrap a single function
  call.

  Args:
    exception_type: The class to wrap exceptions in.
    filename: A filename to use in error messages.

  Yields:
    nothing, just calls the code block.
  """
  try:
    yield
  except utils.UsageError as e:
    raise exception_type(f"Pytype usage error: {utils.message(e)}") from e
  except pyc.CompileError as e:
    raise exception_type("Error reading file %s at line %s: %s" %
                         (filename, e.lineno, e.error)) from e
  except libcst.ParserSyntaxError as e:
    # TODO(rechen): We can get rid of this branch once we delete
    # directors.parser_libcst.
    raise exception_type("Error reading file %s at line %s: %s" %
                         (filename, e.raw_line, e.message)) from e
  except SyntaxError as e:
    raise exception_type("Error reading file %s at line %s: %s" %
                         (filename, e.lineno, e.msg)) from e
  except directors.SkipFileError as e:
    raise exception_type("Pytype could not analyze file %s: "
                         "'# skip-file' directive found" % filename) from e
  except pytd_utils.LoadPickleError as e:
    raise exception_type(f"Error analyzing file {filename}: Could not load "
                         f"serialized dependency {e.filename}") from e
  except Exception as e:  # pylint: disable=broad-except
    msg = f"Pytype error: {e.__class__.__name__}: {e.args[0]}"
    raise exception_type(msg).with_traceback(e.__traceback__)
